import os
import json
import requests

from ..logger import log
from ..colors import colorLib

def resolve(args, signature, type="signatures"):
  try:
    sigRequest = requests.get(f'https://www.4byte.directory/api/v1/{type}/?format=json&hex_signature=0x{signature}', timeout=3)
    if sigRequest.status_code == 200:
      sigBody = json.loads(sigRequest.text)
      if len(sigBody['results']) > 0:
        
        # parse the returned signatures, there can be multiple so we need to store it in a potential list of resolved signatures.
        textSignatures = []
        for sig in sigBody['results']:
          textSignatures.append(sig["text_signature"])

        return textSignatures
      
    else:
      return None
  except:
    return None



from hashlib import md5
import json
import os
import importlib
import pathlib
import traceback

from heimdall.lib.utils.io import write

from ..utils.colors import colorLib
from ..utils.logger import log, logTraceback

def getModules(args=None):
  mods = []
  dist = []
  max_title_length = 0
  max_description_length = 0
  
  # get all modules in heimdall/lib/modules
  for item in os.listdir(pathlib.Path(__file__).parent.resolve()):
    if item.endswith('.py') and not "modules" in item and not "__init__" in item:
      try:
        # import the module to get the meta data
        temp = importlib.import_module(f'.lib.modules.{item.replace(".py", "")}', package='heimdall')
        meta = temp.meta
        meta['import'] = f'.lib.modules.{item.replace(".py", "")}'
        
        # store max lens for help menu formatting
        if len(meta['title']) >= max_title_length:
          max_title_length = len(meta['title'])
        if len(meta['description']) >= max_description_length:
          max_description_length = len(meta['description'])
          
        # add the module to the dist
        mods.append(meta)
        dist.append({
          "name": meta['title'],
          "version": meta['version'],
          "checksum": md5(open(f'{pathlib.Path(__file__).parent.resolve()}/{item}' ,'rb').read()).hexdigest()
        })
      except Exception as e:
        
        # if a module fails to mount, log the error and continue to other modules
        log('warning', f'Module {colorLib.YELLOW}{item}{colorLib.RESET} failed to mount!')
        logTraceback(traceback.format_exc(), True)
  
  # write the hashsums to the dist file, will be used when we have a package manager
  write(f'{pathlib.Path(__file__).parent.resolve()}/dist.json', json.dumps(dist, indent=4))
  return (mods, max_title_length, max_description_length)


import json
import os
import hashlib
import pickle
from .logger import log
from .colors import colorLib

# gets the output directory of a given run
def outputDirectory(outpath, target):
  if outpath:
    output = f'{outpath}/{target}/'
  else:
    output = f'{os.getcwd()}/output/{target}'
  
  # if the path doesnt exist, make it
  if not os.path.exists(output):
    log('info', f'Creating output directory {colorLib.CYAN}{output.replace(os.getcwd(), ".")}{colorLib.RESET}')
    os.makedirs(output)
  
  return output

# write to a file
def write(file, lines):
  with open(file,'w') as f:
    f.write(lines)

# delete a file if it exists
def delete(file):
  if os.path.exists(file):
    os.remove(file)

# write an object to a file as a pickle object
def writeObj(file, lines):
  with open(file,'wb') as f:
    pickle.dump(lines, f)

# make a path if it doesnt exist
def makePath(dirPath):
  if not os.path.exists(dirPath):
    log('info', f'Creating source directory {colorLib.CYAN}{dirPath.replace(os.getcwd(), ".")}{colorLib.RESET}')
    os.makedirs(dirPath)
  return dirPath

# append a single line to a file
def appendLine(file, line):
  if not os.path.exists(file):
    with open(file,'w') as f:
      f.write(line)
  else:
    with open(file,'a') as f:
      f.write(line)

# returns whether or not a file exists
def pathExists(dirPath):
  return os.path.exists(dirPath)

# reads lines from a file path
def readFile(path):
  with open(path) as f:
    lines = f.readlines()
  return lines

# calculates a directories checksum by calculating the ,d5 hash of all files in the directory
def checksum(directory, result=None):
  result = result if result else ''
  for item in os.listdir(directory):
    if item.endswith("py"):
      result = hashlib.md5((str(open(f'{directory}/{item}','rb').read()) + result).encode('utf-8')).hexdigest()
    elif "." not in item:
      result = checksum(f'{directory}/{item}', result)
      
  return result

# writes a json object to the config file after changing pair {name: value}
def write_config_value_at_path(path, name, value):
    with open(path, 'r') as f:
        config = json.load(f)
        config[name] = value
    with open(path, 'w') as f:
        json.dump(config, f, indent=4)
        
        
# load a file as  a json object
def loadFileAsJson(path):
  with open(path) as pathFile:
    contents = json.loads("".join(pathFile.readlines()))
  return contents

# load a file as a pickle object
def loadFileAsPickle(path):
  with open(path, 'rb') as pathFile:
    contents = pickle.load(pathFile)
  return contents


import os
import re
import json
import traceback
import multiprocessing

from bidict import bidict

from ..io import *
from ..colors import colorLib
from ..apis.sigdir import resolve
from ..eth.classes.vm import VirtualMachine
from ..eth.classes.function import Function
from ..logger import log, logTraceback, progress_bar, logfile


# gets the history of a trace, -ith index
def traceHistory(trace, _i=0):
  return trace[-1*(_i+1)]

# finds all signatures by looking through the EVM dispatcher
def findSignatures(assembly, trace, args):
  signatures = []
  
  # finding the end of the function dispatcher
  for i in range(len(trace['trace'])):
    historical = traceHistory(trace['trace'], i)
    if historical['opcode'] == "JUMPI":
      endDispatcher = historical['inputs'][0]
      break
  
  # finding all PUSH4 instructions which occur before the end of the dispatcher
  for key in assembly:
    if key <= endDispatcher:
      if assembly[key]['opcode']['name'] == "PUSH4":
        if not assembly[key]['argument'] in signatures:
         signatures.append(assembly[key]['argument'])

  return [signatures, endDispatcher]

# main function for resolving functions
def resolveFunctions(assembly, args, output):
  try:
    functions = []
    log('info', f'Tracing EVM to determine function signatures...', True)
    
    # create a new VM instance with the disassembled target bytecode
    dispatcher = VirtualMachine(_assembly=assembly)
    dispatcherTrace = dispatcher.trace()
    
    # calculate the end of the dispatcher and all function signatures
    [signatures, endDispatcher] = findSignatures(assembly, dispatcherTrace, args)
    indicesDict = {}
    potentialDict = {}

    log('info', f'Found {colorLib.CYAN}{len(signatures)}{colorLib.RESET} unique signatures. Parsing functions...', True)
    log('info', f'Determining function entry points and resolving signatures...', True)
  
  except:
    
    # if theres an exception, we cant do anything since there were no found signatures. It could be a very
    # old compiler version or a compiler which doesnt use a dispatcher like Solidity, ex: Vyper.
    log('critical', f'Tracing EVM failed! Advanced logs available at {colorLib.RED + colorLib.UNDERLINE + logfile + colorLib.RESET}.')
    logTraceback(traceback.format_exc(), True)
    return

  # if the user wants to flush the cache, or if the cache doesnt exist, trace the signatures and cache some info about them
  if args.flush or args.ignore_cache or not pathExists(f'{output}/__cache__/indices.pickle') or not pathExists(f'{output}/__cache__/signatures.pickle'):
    
    # create a new function object with the signature, assembly, and the PC of the end of the dispatcher.
    # this will return a list of potential resolved function names
    for sig in progress_bar(signatures, args):
      function = Function(args, sig, assembly, endDispatcher)
      indicesDict[function.index] = sig
      potentialDict[sig] = function.potentialNames

    # write the cache fules
    makePath(f'{output}/__cache__')
    writeObj(f'{output}/__cache__/indices.pickle', indicesDict)
    writeObj(f'{output}/__cache__/signatures.pickle', potentialDict)
  else:
    
    # else, load the cache and save time
    log('info', 'Loaded signatures from cache!', True)
    indicesDict = loadFileAsPickle(f'{output}/__cache__/indices.pickle')
    potentialDict = loadFileAsPickle(f'{output}/__cache__/signatures.pickle')

  # convert the 1D indicesDict to a bidict
  indices = bidict(indicesDict)
  events = []
  mappings = []

  # for each function signature, create a new function object with some more information from the cache / resolved values
  # each object will be appended to a list of function objects for writing
  log('info', f'Determining function parameters, views, and returns...', True)
  for sig in progress_bar(signatures, args):
    try:
      function = Function(args, sig, assembly, endDispatcher, indices.inverse[sig], indices, potentialDict[sig])
      functions.append(function)
    except KeyError as e:
      
      # the signature failed to be resolved or found in cache
      pass
    except Exception as e:
      
      # encountered an unexpected error, log it and continue
      log('info', f'Ignoring signature {colorLib.CYAN}{hex(sig)}{colorLib.RESET}. Trace execution excepted!', True)
      logTraceback(traceback.format_exc(), True)
  
  # build the log output and save all mappings / event logs to a list for writing
  functionLogString = f'Determined return, typing, and parameters.'
  for i, f in enumerate(functions):
    functionLogString += (f'\n{" "*25}{"├" if i+1 < len(functions) else "└"}─({colorLib.CYAN}{i}{colorLib.RESET}) '
        f'{colorLib.CYAN}{f.name}{f.params} '
        f'{colorLib.RESET}{"external " if f.external else "public "}'
        f'{colorLib.RESET}{"" if f.external else "pure " if f.pure else "view " if f.view else ""}'
        f'{colorLib.RESET}{"payable " if f.payable else ""}'
        f'{f"returns({f.returns})" if f.returns else ""}'
        )
    
    # check if each event is unique, and insert it to the event list if it is
    for event in f.events:
      if not any(e['signature'] == event['signature'] for e in events):
        events.append(event)
      
    # check if each mapping is unique, and insert it to the mapping list if it is
    for mapping in f.mappings:
      if not any(m['slot'] == f.mappings[mapping]['slot'] for m in mappings):
        mappings.append(f.mappings[mapping])
  
  # log some info about the functions
  log('info', functionLogString, True)
  if all(func.view == True for func in functions):
    log('warning', 'No non-view functions found. Is this a proxy contract?', True)
  log('info', f'Found {colorLib.CYAN}{len(events)}{colorLib.RESET} unique events. Resolving signatures...', True)
  
  # if we are ignoring the cache or the cache doesnt exist, write the event logs to the cache
  if args.flush or args.ignore_cache or not pathExists(f'{output}/__cache__/events.pickle'):
    eventDict = {}
    
    # for each event, try to resolve it's signature on 4byte.directory
    for event in progress_bar(events, args):
      resolvedEvent = resolve(args, event['signature'][2:], 'event-signatures')
      if resolvedEvent:
        eventParams = re.search(r'\((.*)\)', resolvedEvent[0]).group(1).split(",")
        if '' in eventParams:
          eventParams.remove('')
        eventName = resolvedEvent[0].split("(")[0]
        eventDict[event['signature']] = {"name": eventName, "params": eventParams}
      else:
        
        # no match found, use a generic name. ex: event_ffffffff
        eventDict[event['signature']] = {"name": f'event_{event["signature"][2:10]}', "params": ["bytes" for i in range(event['topicCount']-1)]}
    
    # write the events to the cache
    writeObj(f'{output}/__cache__/events.pickle', eventDict)
    log('info', f'Resolved signature names.', True)
  else:
    
    # the cache file exists, load the events
    eventDict = loadFileAsPickle(f'{output}/__cache__/events.pickle')
  
  # building the abi
  abi = []
  abi.append({
      "type": "constructor",
      "inputs": [],
      "stateMutability": "nonpayable"
    })
  
  # write events to the abi
  for key in eventDict:
    abi.append({
      "type": "event",
      "name": eventDict[key]['name'],
      "inputs": [{
        "name": f'arg{index}',
        "internalType": input,
        "type": input
        } for index, input in enumerate(eventDict[key]['params'])],
    })
  
  # write functions to the abi, saving each constant storage value to a list
  constantStorage = {}
  for func in functions:
    if func.constant != False:
      constantStorage[func.name.replace('func_', 'storage_')] = func.constant
    abi.append({
      "type": "function",
      
      # if the function is a storage getter, change the default name to be more descriptive
      "name": func.name if not func.constant else func.name.replace('func_', 'storage_'),
      "inputs": [{
        "name": f'arg{index}',
        "internalType": input,
        "type": input
        } for index, input in enumerate(func.params)],
      "outputs": [{
        "name": f'ret0',
        "internalType": func.returns.replace(" memory", ""),
        "type": func.returns.replace(" memory", ""),
        }] if func.returns else [],
      "stateMutability": "payable" if func.payable else 
                          "nonpayable" if func.external else
                          "pure" if func.pure else 
                          "view" if func.view else 
                          "nonpayable",
      "constant": True if func.constant != False else False
    })
    
  # write the abi to the output directory
  write(f'{output}/abi.json', json.dumps(abi, indent=2))
  log('success', f'Wrote ABI file {colorLib.GREEN}{output.replace(os.getcwd(), ".")}/abi.json{colorLib.RESET}')
    
  return [abi, functions, bidict(indicesDict), signatures, eventDict, constantStorage, mappings]


import re
import traceback

from heimdall.lib.utils.logger import logTraceback

from .logic import Logic


def postProcess(_line, signatures, events, constantStorage):
  
  # cleaning up logic that doesn't actually do anything
  _line = re.sub(r'( << 1 - 1|\/ 1|0 + | + 0)', '', _line)

  # cleaning up redundant castings
  
  try:
    cleaned = _line
    
    # replace all Event_0x00000000 with resolved event name
    eventPlaceholder = re.findall(r'Event_0x[a-fA-F0-9]{8}', cleaned, re.IGNORECASE)
    if len(eventPlaceholder) >= 1:
      placeholderSignature = eventPlaceholder[0][6:16]
      for eventSignature in events:
        if eventSignature.startswith(placeholderSignature):
          cleaned = cleaned.replace(eventPlaceholder[0], events[eventSignature]['name'])
    
    # replace all constant SLOADS with their names
    storage = re.findall(r'LOAD\{[0-9]{0,3}\}', cleaned, re.IGNORECASE)
    for i, access in enumerate(storage):
      access = access.replace("LOAD{", "").replace("}", "")
      handled = False
      for key in constantStorage:
        if constantStorage[key] == int(access):
          cleaned = cleaned.replace(storage[i], f'_{key}')
          handled = True
          break
      if not handled:
        cleaned = cleaned.replace(storage[i], f'_storage_{hex(int(access))}')
        
  except Exception as e:
    logTraceback(traceback.format_exc(), True)
  return cleaned




from copy import deepcopy
import re
import traceback
from web3 import Web3

from ...logic import Any, Logic, _match, bytesToType, commonTypes, determineType, solidify_wrapped, offsetToMemoryName, solidify
from ...apis.sigdir import resolve
from ...eth.classes.vm import VirtualMachine
from ...eth.classes.stack import Stack
from ...logger import log, logTraceback, query
from ...colors import colorLib

class Function():
  
  # initialize a new function object with some default information
  def __init__(self, args, signature, assembly, endDispatcher, index=None, indices=None, potentialNames=None, params=None):
    self.args = args
    self.event_signatures = []
    self.events = []
    self.signature = Logic.padHex(None, signature, 8)
    self.name = None
    self.assembly = assembly
    self.potentialNames = potentialNames if potentialNames != None else resolve(args, self.signature[2:])
    self.endDispatcher = endDispatcher
    self.index = index if index != None else self.getEntryPoint()
    self.indices = indices
    self.returns = None
    self.external = False
    self.view = True
    self.pure = True
    self.payable = True
    self.constant = False
    self.logic = []
    self.warnings = []    
    self.memory = {}
    self.memlast = {}
    self.mappings = {}
    self.storage = None
    
    if self.indices != None:
      self.params = params if params != None else self.resolveParams()
  
  # determines if the jumpdest is a jump to another function
  # ( I think this could be somehow used to find internal functions, dont quote me on that )
  def isFunctionCall(self, dest):
    return dest if (dest in [item for sublist in self.indices.keys() for item in sublist]) else None

  # determines the entry point of the function
  def getEntryPoint(self):
    
    # create a new VM with the disassembly of the target bytecode
    vm = deepcopy(VirtualMachine(_assembly=self.assembly, _calldata=int(self.signature, 16)))
   
   # recursively make a tree of all jumpdests and their conditions
    def resolve(entries=[], logJump=False):
      try:
        
        # get the next instruction and execute it on the VM
        instructionIndex = list(self.assembly.keys()).index(vm.instruction)
        code = self.assembly[vm.instruction]
        lastPC = vm.instruction
        vm.lastInstruction = lastPC
        vm.instruction = (list(self.assembly.keys())[instructionIndex+1])
        if vm.instruction > self.endDispatcher:
          return tuple(entries)
        vm.execute(code['opcode']['name'], code['argument'])
        call = vm.stacktrace[-1]
        
        # if the instruction is a push4 and the value matches the signature, we have found an entry point  
        if call['opcode'] == "PUSH4" and call['outputs'] == int(self.signature, 16):
          return resolve(entries, True)
        
        # if we already found an entry point, get the value of the jumpdest. This is the entry point for the function
        if logJump and code['opcode']['name'] in ['JUMP', 'JUMPI']:
          if call['outputs']:
            entries.append(call['outputs'])
          else:
            if call['inputs']:
              entries.append(call['inputs'][0])
            else:
              entries.append(vm.stacktrace[-2]['outputs'])
          return resolve(entries, False)
        return resolve(entries, logJump)
      
      except Exception as e:
        log('warning', 'Exception encountered when resolving function index.', not self.args.verbose)
        logTraceback(traceback.format_exc(), True)
        
    return resolve()
  
  
  def resolveParams(self):
    def trace(calldata, startPoint=1, stack=Stack(), stacktrace=[], handled=[]):
      
      # create a new VM object with a set of arguments which can come from the recursive tree
      vm = deepcopy(VirtualMachine(_assembly=self.assembly, _calldata=calldata, _instruction=startPoint, _stack=stack, _stacktrace=stacktrace))
      
      def map(mapping=[], getMask=False):
        
        # get the next instruction and execute it on the VM
        find_cd_mask = getMask
        instructionIndex = list(self.assembly.keys()).index(vm.instruction)
        code = self.assembly[vm.instruction]
        lastPC = vm.instruction
        vm.lastInstruction = lastPC
        vm.instruction = (list(self.assembly.keys())[instructionIndex+1])
        next = (list(self.assembly.keys())[instructionIndex+1])
        ret = vm.execute(code['opcode']['name'], code['argument'])
        call = vm.stacktrace[-1]
        
        # if the instruction executed after the dispatcher, it's part of the function logic 
        if call['pc'] > self.endDispatcher:
          
          # if the instruction is a state reading instruction, it's no longer a pure function
          if call['opcode'] in ['SSTORE', 'SLOAD', 'DELEGATECALL', 'STATICCALL', 'CALL', 'CALLCODE']:
            self.pure = False
            
            # convert all SLOAD operations to a value and store it
            if call['opcode'] == "SLOAD":
              self.storage = solidify_wrapped(call["wrapped"], vm, self)
            
            # if we are storing a value, add the store logic to the decompiled contract
            elif call['opcode'] == "SSTORE":
              key = solidify_wrapped(("SLOAD", Any, call['wrapped'][0]), vm, self)
              value = solidify_wrapped(call["wrapped"][1:], vm, self)
                  
              # appending logic
              self.logic.append([call['pc'], f'{key} = {value};'])
              self.view = False

            else:
              
              # handling for external contract calls
              callParameters = call['inputs']
              # if delegatecall or staticcall, insert an empty value field
              if call['opcode'] in ('STATICCALL', 'DELEGATECALL'):
                callParameters.insert(2, 0)
                
              # get and add all sources for the external call to the solidity logic
              source = solidify_wrapped(call['wrapped'][1], vm, self)
              self.logic.append([call['pc'], f'bytes memory {source} = {self.memory[source]["value"]};'])
              
              # I'm sorry this is extremely ugly, we just have to actually parse these 
              # callParameters into a solidity staticcall, and I dont want a massive fstring
              # so I broke it up a bit into parts
              self.logic.append([call['pc'] ,(
                f'(bool success, bytes{callParameters[6]} memory {offsetToMemoryName(callParameters[5])}) =' 
                f' address({source}).staticcall'
                f'({offsetToMemoryName(callParameters[3])});'
              )])
              
              # this is no longer a view function since it modifies the state
              self.view = False
          
          # if the instruction is an external call, add an external tag to this function object
          if call['opcode'] in ['DELEGATECALL', 'STATICCALL', 'CALL']:
            self.external = True

          # save all memory storages to the decompiled contract
          if call['opcode'] in ['MSTORE', 'MSTORE8']:
            try:
              self.memlast[offsetToMemoryName(call["inputs"][0])] = self.memory[offsetToMemoryName(call["inputs"][0])]
            except:
              pass
            self.memory[offsetToMemoryName(call["inputs"][0])] = {
              "value": solidify_wrapped(call["wrapped"][1:], vm, self),
              "pc": call['pc']
            }

          # add all event emissions to the decompiled contract
          if 'LOG' in call['opcode']:
            
            # add all sources for the external call to the solidity logic
            sources = []
            for temp in call['wrapped']:
              sources.append(str(solidify_wrapped(temp, vm, self)))
            sources += call['event']['topics']['unindexed']
            [self.logic.append([call['pc'], f'bytes memory {source} = {self.memory[source]["value"]};']) for source in sources if re.match(r"var[0-9]*", source)]
            
            # add the event to the decompiled contract
            self.logic.append([call['pc'], f'emit Event_{call["event"]["signature"][:10]}({", ".join(sources)});'])

          # logic for calldataloads and determining function types
          if 'wrapped' in call and call['wrapped'] != None:
            
            # if the opcode is an AND, it's typically a masking operation and we can calculate the type with the mask value
            if call['opcode'] == "AND" and (
              (m := _match(call['wrapped'], [(':offset', 0, 'CALLDATALOAD'), (None, ':mask', Any)])) or 
              (m := _match(call['wrapped'], [(None, ':mask', Any), (':offset', 0, 'CALLDATALOAD')]))
            ):
              mapping.append([ (len(hex(m.mask)[2:])*4) , m.offset ])
            
            # if the opcode is a ISZERO, it's typically a boolean
            elif call['opcode'] == 'ISZERO':
              if(m := _match(call['wrapped'], [(':offset', 0, 'CALLDATALOAD')])):
                mapping.append([ -1 , m.offset ])
            
            # If we are adding 4 to a calldataload, it's typically a pointer to somewhere in memory or storage. This is an indicator of an array as calldata
            elif call['opcode'] == "ADD" and (
              (m := _match(call['wrapped'], [(None, 4, 'PUSH1'), (':offset', 0, 'CALLDATALOAD')])) or 
              (m := _match(call['wrapped'], [(':offset', 0, 'CALLDATALOAD'), (None, 4, 'PUSH1')]))
            ):
              mapping.append([ -2 , m.offset ])
            
            # Default to common calldata types
            elif call['opcode'] == "CALLDATALOAD":
              mapping.append([ 256, call['inputs'][0] ])
          
          # the meat of the recursive logic happens here
          if code['opcode']['name'] == "JUMPI":
            if call['inputs']:
              jumpdest = call['inputs'][0]
              
              # if the call is a conditional jump and the condition was true:
              # - create a new VM with the next PC as if this jump wasnt taken
              # - run the current VM with the next PC as if this jump was taken
              #
              # eventually loop detection will reside here. For now, we break out of loops by checking if the next PC was already handled.
              if call['inputs'][1] == 1 and next not in handled:
                handled.append(next)
                mapping += trace(calldata, next, deepcopy(vm.stack), stacktrace=[call], handled=handled)
                
              # the jump conditional wasn't true:
              # - create a new VM with the next PC as if this jump was taken
              # - run the current VM with the next PC as if this jump wasn't taken
              elif jumpdest not in handled:
                handled.append(jumpdest)
                mapping += trace(calldata, jumpdest, deepcopy(vm.stack), stacktrace=[call], handled=handled)
              
              else:
                
                # return with a loop breakout code
                ret = (5, 0)

        # if the current VM execution returned a value:
        if ret:
          
          # save all events from the logs to this function's event list
          for event in vm.logs:
            if event['signature'] not in self.event_signatures:
              self.event_signatures.append(event['signature'])
              self.events.append(event)
          
          # determine the type of the return value
          if ret[0] == 0:
            if self.pure: self.pure = ret[1]
            if self.returns in commonTypes(): self.returns = determineType(self, ret[1])
          
          # the execution REVERTED
          elif ret[0] == 1:
            try:
              revert_string = bytesToType(['string'], ret[1])[0]
              if not '\x00' in revert_string:
                
                # trace backwards through the stacktrace to find the call that caused the revert
                for call in reversed(vm.stacktrace):
                  if call['opcode'] == 'JUMPI' and call['wrapped'][1][0]:
                    
                    # calculate the solidity value of the revert reason
                    revert_reason = solidify_wrapped(call["wrapped"][1], vm, self)
                    
                    # payable detection, pretty crude but working as of ^0.8.14.
                    if 'msg.value == 0' in re.sub(r'[\(\)]*', '', revert_reason):
                      self.payable = False
                    
                    # write the require statements with the revert reasons if applicable, skipping internal requires
                    if any(reason in revert_reason for reason in ['arg', 'var', 'mapping', 'memory',]):
                      if revert_string != "0":
                        self.logic.append([call['pc'], f'require({revert_reason}, "{revert_string}");'])
                      else:
                        self.logic.append([call['pc'], f'require({revert_reason});'])
                    break
                  
            except Exception as e:
              logTraceback(traceback.format_exc(), True)
          
          return mapping
        
        return map(mapping, find_cd_mask)
      return map()

    # convert all calldata accesses to unique values
    accesses = trace(int(self.signature, 16))
    uniques = {}
    for cd in accesses:
      if cd[1] not in uniques:
        uniques[cd[1]] = {
          "val": cd[0] if cd[0] != -2 else None,
          "isPointer": cd[0] == -2
        }
      elif uniques[cd[1]]['val'] in (None, 256):
        uniques[cd[1]] = {
          "val": cd[0] if uniques[cd[1]]['val'] == None else cd[0] if cd[0] != -2 else uniques[cd[1]]['val'],
          "isPointer": uniques[cd[1]]['isPointer'] or cd[0] == -2
        }
    
    # convert unique calldata values to potential solidity types
    potentials = []
    for mask in uniques:
      potentials.append(Logic.resolveMask(uniques[mask]))
    
    # if theres no calldataload and it's a view function but not a pure function, 
    # it's probably a getter function for some public variable
    if len(potentials) == 0 and self.view and not self.pure:
      self.constant = self.storage
    
    # if we are a pure function, calculate the return type
    if self.pure and self.pure != True:
      
        retval = str(bytesToType([self.returns], self.pure)[0])
        if self.returns == 'address':
          retval = Web3.toChecksumAddress(retval.lower())
        elif self.returns == 'string':
          retval = f'\'{retval}\''
          
        self.logic.append([0xFFFFFFFFFFFF, f'return {retval};'])
    
    # handle the returns of some other functions. Needs improvement
    elif self.returns == 'bool':
      self.logic.append([0xFFFFFFFFFFFF, f'return true;'])
    
    # handle matching resolved 4byte functions to potential names
    matches = []
    if self.potentialNames:
      for name in self.potentialNames:
        match = True
        try:
          
          # matching all parameter types with potential types found by CALLDATALOAD processing
          resolvedParameters = re.search(r'\((.*)\)', name).group(1).split(",")
          if "" in resolvedParameters:
            resolvedParameters.remove('')
          if len(potentials) == len(resolvedParameters):
            for i, param in enumerate(resolvedParameters):
              match_any = False
              for j, potential in enumerate(potentials):
                if re.sub(r'(\(|\))*', '', param) in potential:
                  match_any = True
                  potential.pop(j)                  
                  break
              if not match_any:
                match = False
                break
          else:
            match = False
                
        except Exception as e:
          
          # no match if the lengths dont match
          if (len(resolvedParameters) != len(potentials)) and (resolvedParameters[0] != ''):
            match = False
        
        # add match to a list of potential matches  
        if match:
          matches.append(name)
          
    # ask the user to pick a potential function name if there are multiple matches
    if len(matches) > 1:
      logString = 'Multiple function possibilities found:'
      for i, match in enumerate(matches):
        logString += f'\n{" "*25}{"├" if i+1 < len(matches) else "└"}─({colorLib.CYAN}{i}{colorLib.RESET}) {colorLib.CYAN}{match}{colorLib.RESET}'
      
      # if the user has default specified, use the oldest potential name. These tend to be more accurate
      if not self.args.default:
        logString += f'\n\n{" "*25}Select one of the above matches for this function [{len(matches)-1}]: '
        selectionRaw = query('info', str(len(matches)-1), logString)
        if selectionRaw.isnumeric():
          selection = int(selectionRaw)
          if selection > len(matches)-1:
            selection = len(matches)-1
        else:
          selection = len(matches)-1
      else:
        selection = len(matches)-1
        logString += f'\n\n{" "*25}Select one of the above matches for this function [{len(matches)-1}]: {len(matches)-1}'
        log('info', logString)
      self.name = matches[selection]
      
    elif len(matches) > 0:
      self.name = matches[0]

    # build the function's selected calldata types
    ret = []
    if self.name:
      try:
        ret = re.search(r'\((.{1,})\)', self.name).group(1).split(",")
      except: pass
    else: 
      for param in potentials:
        if len(param) > 0:
          ret.append(param[0])
    
    # if theres a resolved name, use it over the placeholder name ( ex: func_ffffffff )
    if self.name:
      self.name = self.name.split("(")[0]
    else:
      self.name = f'func_{self.signature}'

    return tuple(ret)

  # python3 heimdall.py -m 1 -v --default -t 0x4Ae258F6616Fc972aF90b60bFfc598c140C79def
  
  # TODO: re-add internal funcition calls from self.indices


from curses import has_key
import math
import traceback
import rlp
import copy

from web3 import Web3
web3 = Web3(Web3.HTTPProvider("https://mainnet.infura.io/v3/422705d85b5c4923b841a60ba0c01129"))


from ...logger import log
from ...colors import colorLib
from ...logic import Logic, offsetToMemoryName
from ...eth.classes.stack import Stack
from ...eth.classes.memory import Memory

from Crypto.Hash import keccak
sha3 = lambda x: keccak.new(digest_bits=256, data=x).digest()

class VirtualMachine:
  
  # initialize the virtual machine with a set of options for initial EVM state
  def __init__(self, 
    _stack = Stack(),
    _memory= Memory(), 
    _storage={}, 
    _calldata=0,
    _address=0, 
    _origin=0, 
    _caller=0, 
    _callValue=0, 
    _bytecode=0, 
    _returnData=0,
    _assembly={},
    _instruction=1,
    _gas=1000000,
    _stacktrace=[],
    _logEvents=False):
    self.stack = _stack
    self.memory = _memory
    self.storage = _storage
    self.calldata = _calldata
    self.address = _address
    self.caller = _caller
    self.origin = _origin
    self.value = _callValue
    self.bytecode = _bytecode
    self.returnData = _returnData
    self.assembly = _assembly
    self.gas = _gas
    self.instruction = _instruction
    self.lastInstruction = _instruction
    self.logs = []
    self.stacktrace = _stacktrace
    self.logic = Logic()
    self.logEvents = _logEvents
    
  # convert any hex to a checksummed address
  def toAddress(self, hex):
    return Web3.toChecksumAddress(self.logic.padHex(hex, 40))
  
  # appends an event to the logs
  def logEvent(self, event):
    self.logs.append(event)
    
  # gets the current EVM stack
  def getStack(self):
    return self.stack.getStack()
  
  # gets the current EVM memory
  def getMemory(self):
    return self.memory
  
  # gets the current EVM storage
  def getStorage(self):
    return self.storage
  
  # gets the last operation performed in the EVM
  def lastOperation(self, n=1):
    try:
      return self.stacktrace[ (n*-1) ]
    except:
      return None
    
  # executes an operation on the EVM 
  def execute(self, opcode, args=None):
    try:
      if type(args) != int:
        args=None

      # pushes a value onto the stack
      if 'PUSH' in opcode:
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "inputs": None,
            "outputs": int(args),
          }
        )
        self.stack.append( int(args), source=opcode )

      # stops execution and returns a value from memory
      elif opcode == 'RETURN':
        [offset_wrapped, size_wrapped] = self.stack.pop(2)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [offset_wrapped, size_wrapped],
            "inputs": [offset_wrapped[1], size_wrapped[1]],
            "outputs": None,
          }
        )
        return (0, f'0x{self.memory.read(offset_wrapped[1], size_wrapped[1]).hex()}')

      # stops execution with an error and returns a value from memory
      elif opcode == 'REVERT':
        [offset_wrapped, size_wrapped] = self.stack.pop(2)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [offset_wrapped, size_wrapped],
            "inputs": [offset_wrapped[1], size_wrapped[1]],
            "outputs": None,
          }
        )
        try:
          ret = f'0x{self.memory.read(offset_wrapped[1], size_wrapped[1]).hex()}'
          return (1, ret)
        except:
          return (1, f'0x')
      
      # stops execution with an error
      elif opcode == 'INVALID':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": (0, 0,),
            "inputs": (0,0,),
            "outputs": None,
          }
        )
        try:
          ret = f'0x{self.memory.read(0,0).hex()}'
          return (2, ret)
        except:
          return (2, f'0x')

      # stops execution and destroys the evm instance
      elif opcode == 'SELFDESTRUCT':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": None,
          }
        )
        return (3, f'0x')

      # sums the top two values on the stack
      elif opcode == 'ADD':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.add(a, b),
          }
        )
        self.stack.append(self.logic.add(a, b), op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # multiplies the top two values on the stack
      elif opcode == 'MUL':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.mul(a, b),
          }
        )
        self.stack.append(self.logic.mul(a, b), op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # subtracts the top two values on the stack
      elif opcode == 'SUB':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.sub(a, b),
          }
        )
        self.stack.append(self.logic.sub(a, b), op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # divides the top two values on the stack
      elif opcode == 'DIV':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.div(a, b),
          }
        )
        self.stack.append(self.logic.div(a,b), op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # signed division of the top two values on the stack
      elif opcode == 'SDIV':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.sdiv(a, b),
          }
        )
        self.stack.append(self.logic.sdiv(a,b), op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # calculates the modulo of the top two values on the stack
      elif opcode == 'MOD':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.mod(a, b),
          }
        )
        self.stack.append(self.logic.mod(a,b), op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # calculates the signed modulo of the top two values on the stack
      elif opcode == 'SMOD':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.smod(a, b),
          }
        )
        self.stack.append(self.logic.smod(a,b), op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # calculates the modulo of the third value on the stack with the sum of the top two values on the stack
      elif opcode == 'ADDMOD':
        [c_wrapped, b_wrapped, a_wrapped] = self.stack.pop(3)
        [c, b, a] = self.stack.unwrap([c_wrapped, b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [c_wrapped, b_wrapped, a_wrapped],
            "inputs": [c, b, a],
            "outputs": self.logic.addmod(a, b, c),
          }
        )
        self.stack.append(self.logic.addmod(a,b,c), op=opcode, source=tuple([c_wrapped, b_wrapped, a_wrapped]))

      # calculates the modulo of the third value on the stack with the product of the top two values on the stack
      elif opcode == 'MULMOD':
        [c_wrapped, b_wrapped, a_wrapped] = self.stack.pop(3)
        [c, b, a] = self.stack.unwrap([c_wrapped, b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [c_wrapped, b_wrapped, a_wrapped],
            "inputs": [c, b, a],
            "outputs": self.logic.mulmod(a, b, c),
          }
        )
        self.stack.append(self.logic.mulmod(a,b,c), op=opcode, source=tuple([c_wrapped, b_wrapped, a_wrapped]))

      # calculates the exponential of the top two values on the stack
      elif opcode == 'EXP':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": self.logic.exp(a, b),
          }
        )
        self.stack.append(self.logic.exp(a,b), op=opcode, source=tuple([b_wrapped, a_wrapped]))
      
      #  extends the sign bit of the top value on the stack
      elif opcode == 'SIGNEXTEND':
        [x_wrapped, b_wrapped] = self.stack.pop(2)
        [x, b] = self.stack.unwrap([x_wrapped, b_wrapped])
        if x < 32:
          tbit = x * 8 + 7
          sbit = 1 << tbit
          if b & sbit:
            self.stacktrace.append(
              {
                "pc": self.lastInstruction,
                "opcode": opcode,
                "wrapped": [x_wrapped, b_wrapped],
                "inputs": [x, b],
                "outputs": b | (self.logic.UINT256_CEILING - sbit),
              }
            )
            self.stack.append( b | (self.logic.UINT256_CEILING - sbit), op=opcode, source=tuple([x_wrapped, b_wrapped]) )
          else:
            self.stacktrace.append(
              {
                "pc": self.lastInstruction,
                "opcode": opcode,
                "wrapped": [x_wrapped, b_wrapped],
                "inputs": [x, b],
                "outputs": b & (sbit - 1),
              }
            )
            self.stack.append( b & (sbit - 1), op=opcode, source=tuple([x_wrapped, b_wrapped])  )
        else:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [x_wrapped, b_wrapped],
              "inputs": [x, b],
              "outputs": b,
            }
          )
          self.stack.append( b, op=opcode, source=tuple([x_wrapped, b_wrapped])  )

      # less than comparison
      elif opcode == 'LT':
        [a_wrapped, b_wrapped] = self.stack.pop(2)
        [a, b] = self.stack.unwrap([a_wrapped, b_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [a_wrapped, b_wrapped],
            "inputs": [a, b],
            "outputs": 1 if a < b else 0,
          }
        )
        self.stack.append(1 if a < b else 0, op=opcode, source=tuple([a_wrapped, b_wrapped]))
      
      # greater than comparison
      elif opcode == 'GT':
        [a_wrapped, b_wrapped] = self.stack.pop(2)
        [a, b] = self.stack.unwrap([a_wrapped, b_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [a_wrapped, b_wrapped],
            "inputs": [a, b],
            "outputs": 1 if a > b else 0,
          }
        )
        self.stack.append(1 if a > b else 0, op=opcode, source=tuple([a_wrapped, b_wrapped]))

      # signed less than comparison
      elif opcode == 'SLT':
        [a_wrapped, b_wrapped] = self.stack.pop(2)
        [a, b] = self.stack.unwrap([a_wrapped, b_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [a_wrapped, b_wrapped],
            "inputs": [a, b],
            "outputs": 1 if a < b else 0,
          }
        )
        self.stack.append(1 if a < b else 0, op=opcode, source=tuple([a_wrapped, b_wrapped]))

      # signed greater than comparison
      elif opcode == 'SGT':
        [a_wrapped, b_wrapped] = self.stack.pop(2)
        [a, b] = self.stack.unwrap([a_wrapped, b_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [a_wrapped, b_wrapped],
            "inputs": [a, b],
            "outputs": 1 if a > b else 0,
          }
        )
        self.stack.append(1 if a > b else 0, op=opcode, source=tuple([a_wrapped, b_wrapped]))

      # equality comparison
      elif opcode == 'EQ':
        [a_wrapped, b_wrapped] = self.stack.pop(2)
        [a, b] = self.stack.unwrap([a_wrapped, b_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [a_wrapped, b_wrapped],
            "inputs": [a, b],
            "outputs": 1 if a == b else 0,
          }
        )
        self.stack.append(1 if b == a else 0, op=opcode, source=tuple([a_wrapped, b_wrapped]))

      # not equal comparison
      elif opcode == 'ISZERO':
        [a_wrapped] = self.stack.pop(1)
        [a] = self.stack.unwrap([a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [a_wrapped],
            "inputs": [a],
            "outputs": 1 if a == 0 else 0,
          }
        )
        self.stack.append(1 if a == 0 else 0, op=opcode, source=tuple([a_wrapped]))

      # and operation
      elif opcode == 'AND':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": a & b,
          }
        )
        self.stack.append(a & b, op=opcode, source=tuple([b_wrapped, a_wrapped]))
    
      # or operation
      elif opcode == 'OR':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": a | b,
          }
        )
        self.stack.append(a | b, op=opcode, source=tuple([b_wrapped, a_wrapped]))
    
      # xor operation
      elif opcode == 'XOR':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [b_wrapped, a_wrapped],
            "inputs": [b, a],
            "outputs": a ^ b,
          }
        )
        self.stack.append(a ^ b, op=opcode, source=tuple([b_wrapped, a_wrapped]))

      # not operation
      elif opcode == 'NOT':
        [a_wrapped] = self.stack.pop(1)
        [a] = self.stack.unwrap([a_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [a_wrapped],
            "inputs": [a],
            "outputs": (~a) & self.logic.UINT256_MAX,
          }
        )
        self.stack.append( (~a) & self.logic.UINT256_MAX, op=opcode, source=tuple([a_wrapped]) )

      # calculates the byte value of a word
      elif opcode == 'BYTE':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        if b >= 32:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [b_wrapped, a_wrapped],
              "inputs": [b, a],
              "outputs": 0,
            }
          )
          self.stack.append( 0, op=opcode, source=tuple([b_wrapped, a_wrapped])  )
        else:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [b_wrapped, a_wrapped],
              "inputs": [b, a],
              "outputs": (a // pow(256, 31 - b)) % 256,
            }
          )
          self.stack.append( (a // pow(256, 31 - b)) % 256, op=opcode, source=tuple([b_wrapped, a_wrapped]) )

      # shifts a word left by a given amount
      elif opcode == 'SHL':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        if b >= 256:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [b_wrapped, a_wrapped],
              "inputs": [b, a],
              "outputs": 0,
            }
          )
          self.stack.append( 0, op=opcode, source=tuple([b_wrapped, a_wrapped]) )
        else:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [b_wrapped, a_wrapped],
              "inputs": [b, a],
              "outputs": ( a << b ) & self.logic.UINT256_MAX,
            }
          )
          self.stack.append( ( a << b ) & self.logic.UINT256_MAX, op=opcode, source=tuple([b_wrapped, a_wrapped]) )
      
      # shifts a word right by a given amount
      elif opcode == 'SHR':
        [b_wrapped, a_wrapped] = self.stack.pop(2)
        [b, a] = self.stack.unwrap([b_wrapped, a_wrapped])
        if b >= 256:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [b_wrapped, a_wrapped],
              "inputs": [b, a],
              "outputs": 0,
            }
          )
          self.stack.append( 0, op=opcode, source=tuple([b_wrapped, a_wrapped]) )
        else:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [b_wrapped, a_wrapped],
              "inputs": [b, a],
              "outputs": ( a >> b ) & self.logic.UINT256_MAX,
            }
          )
          self.stack.append( ( a >> b ) & self.logic.UINT256_MAX, op=opcode, source=tuple([b_wrapped, a_wrapped]) )

      # calculates the keccak256 hash of a word
      elif opcode == 'SHA3':
        [offset_wrapped, size_wrapped] = self.stack.pop(2)
        [offset, size] = self.stack.unwrap([offset_wrapped, size_wrapped])
        mem = self.memory.read(offset,size)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [offset_wrapped, size_wrapped],
            "inputs": [offset, size],
            "memory": mem.hex(),
            "outputs": int(sha3(mem).hex(),16),
          }
        )
        self.stack.append(int(sha3(mem).hex(),16), op=opcode, source=tuple([offset_wrapped, size_wrapped]))

      # returns the address of the EVM contract
      elif opcode == 'ADDRESS':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": self.address,
          }
        )
        self.stack.append( self.address, source=opcode )

      # returns the balance of the EVM contract
      elif opcode == 'BALANCE':
        [a_wrapped] = self.stack.pop(1)
        [a] = self.stack.unwrap([a_wrapped])
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [a_wrapped],
              "inputs": [a],
              "outputs": web3.eth.get_balance( self.toAddress(a) ),
            }
          )
          self.stack.append(web3.eth.get_balance( self.toAddress(a) ), op=opcode, source=tuple([a_wrapped]))
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [a_wrapped],
              "inputs": [a],
              "outputs": 0,
            }
          )
          self.stack.append( 0, op=opcode, source=tuple([a_wrapped]) )

      # returns the origin of the EVM transaction
      elif opcode == 'ORIGIN':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": self.origin,
          }
        )
        self.stack.append( self.origin, source=opcode )

      # returns the caller of the EVM transaction
      elif opcode == 'CALLER':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": self.caller,
          }
        )
        self.stack.append( self.caller, source=opcode )

      # returns the call value of the EVM transaction in wei
      elif opcode == 'CALLVALUE':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": self.value,
          }
        )
        self.stack.append( self.value, source=opcode )

      # gets the calldata of the EVM transaction at ith offset and a size of 32 bytes
      elif opcode == 'CALLDATALOAD':
        [i_wrapped] = self.stack.pop(1)
        [i] = self.stack.unwrap([i_wrapped])
        if len(hex(self.calldata))%2 == 1:
          safeHex = self.logic.padHex(self.calldata, len(hex(self.calldata)[2:])+1)
        else:
          safeHex = hex(self.calldata)
        callData = bytearray.fromhex(safeHex[2:len(safeHex)-(i*2)])[0:32]
        while len(callData) < 32:
          callData += bytes(1)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [i_wrapped],
            "inputs": [i],
            "outputs": int(callData.hex(), 16),
          }
        )
        self.stack.append ( int(callData.hex(), 16), source=opcode, op=i)

      # gets the size of the calldata of the EVM transaction
      elif opcode == 'CALLDATASIZE':
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": self.logic.byteSize(self.calldata),
            }
          )
          self.stack.append( self.logic.byteSize(self.calldata), source=opcode )
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # copies the calldata to memory
      elif opcode == 'CALLDATACOPY':
        [destOffset_wrapped, offset_wrapped, size_wrapped] = self.stack.pop(3)
        [destOffset, offset, size] = self.stack.unwrap([destOffset_wrapped, offset_wrapped, size_wrapped])

        # if the value isnt even, prepend a 0
        if len(hex(self.calldata))%2 == 1:
          safeHex = self.logic.padHex(self.calldata, len(hex(self.calldata)[2:])+1)
        else:
          safeHex = hex(self.calldata)
 
        # extend memory by the calldatasize
        extended = bytearray.fromhex(safeHex[2:len(safeHex)-(offset*2)])
        self.memory.extend(destOffset, size)
        self.memory.write(destOffset, size, extended)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [destOffset_wrapped, offset_wrapped, size_wrapped],
            "inputs": [destOffset, offset, size],
            "outputs": None,
          }
        )

      # gets the bytecode size of a contract
      elif opcode == 'CODESIZE':
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": self.logic.byteSize(self.bytecode),
            }
          )
          self.stack.append( self.logic.byteSize(self.bytecode), source=opcode )
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # copies the bytecode to memory
      elif opcode == 'CODECOPY':
        [destOffset_wrapped, offset_wrapped, size_wrapped] = self.stack.pop(3)
        [destOffset, offset, size] = self.stack.unwrap([destOffset_wrapped, offset_wrapped, size_wrapped])
        
        # extend memory by the codesize and copy the bytecode to memory
        extended = bytearray.fromhex(hex(self.bytecode)[2:len(hex(self.bytecode))-(offset*2)])
        self.memory.extend(destOffset, size)
        self.memory.write(destOffset, size, extended)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [destOffset_wrapped, offset_wrapped, size_wrapped],
            "inputs": [destOffset, offset, size],
            "outputs": None,
          }
        )
      
      # gets the current gas price in wei
      elif opcode == 'GASPRICE':
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": Web3.toWei(20, 'gwei'),
            }
          )
          self.stack.append(Web3.toWei(20, 'gwei'), source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # gets the codesize of an external account
      elif opcode == 'EXTCODESIZE':
        [target_wrapped] = self.stack.pop(1)
        [target] = self.stack.unwrap([target_wrapped])
        try:
          rawBytecode = web3.eth.get_code(self.toAddress(target))
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [target_wrapped],
              "inputs": [target],
              "outputs": self.logic.byteSize(rawBytecode),
            }
          )
          self.stack.append( self.logic.byteSize(rawBytecode), op=opcode, source=tuple([target_wrapped]) )
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [target_wrapped],
              "inputs": [target],
              "outputs": 0,
            }
          )
          self.stack.append(0, op=opcode, source=tuple([target]))

      # copies the external account's bytecode to memory
      elif opcode == 'EXTCODECOPY':
        [address_wrapped, destOffset_wrapped, offset_wrapped, size_wrapped] = self.stack.pop(3)
        [address, destOffset, offset, size] = self.stack.unwrap([address_wrapped, destOffset_wrapped, offset_wrapped, size_wrapped])
        try:
          rawBytecode = web3.eth.get_code(self.toAddress(address))

          # extend memory by the codesize and copy the bytecode to memory
          extended = bytearray.fromhex(hex(rawBytecode)[:len(hex(rawBytecode))-(offset*2)])
          self.memory.extend(destOffset, size)
          self.memory.write(destOffset, size, extended)
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [address_wrapped, destOffset_wrapped, offset_wrapped, size_wrapped],
              "inputs": [address, destOffset, offset, size],
              "outputs": None,
            }
          )
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [address_wrapped, destOffset_wrapped, offset_wrapped, size_wrapped],
              "inputs": [address, destOffset, offset, size],
              "outputs": None,
            }
          )

      # calculate the returndata size of a call
      elif opcode == 'RETURNDATASIZE':
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": self.logic.byteSize(self.returnData),
            }
          )
          self.stack.append( self.logic.byteSize(self.returnData), source=opcode )
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # copy the return data to memory
      elif opcode == 'RETURNDATACOPY':
        [destOffset_wrapped, offset_wrapped, size_wrapped] = self.stack.pop(3)
        [destOffset, offset, size] = self.stack.unwrap([destOffset_wrapped, offset_wrapped, size_wrapped])
        
        # pad the return data with zeros until the length is even
        if len(hex(self.returnData))%2 == 1:
          safeHex = self.logic.padHex(self.returnData, len(hex(self.returnData)[2:])+1)
        else:
          safeHex = hex(self.returnData)

        # extend the memory and copy the return data to memory
        extended = bytearray.fromhex(safeHex[2:len(safeHex)-(offset*2)])
        self.memory.extend(destOffset, size)
        self.memory.write(destOffset, size, extended)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
             "wrapped": [destOffset_wrapped, offset_wrapped, size_wrapped],
            "inputs": [destOffset, offset, size],
            "outputs": None,
          }
        )

      # get the hash of the code of an external account
      elif opcode == 'EXTCODEHASH':
        [target_wrapped] = self.stack.pop(1)
        [target] = self.stack.unwrap([target_wrapped])
        try:
          rawBytecode = web3.eth.get_code(self.toAddress(target))
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [target_wrapped],
              "inputs": [target],
              "outputs": int(Web3.keccak(rawBytecode).hex(),16),
            }
          )
          self.stack.append( int(Web3.keccak(rawBytecode).hex(),16), op=opcode, source=tuple([target_wrapped]) )
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [target_wrapped],
              "inputs": [target],
              "outputs": 0,
            }
          )
          self.stack.append(0, op=opcode, source=tuple([target_wrapped]))

      # get the hash of the a block 
      elif opcode == 'BLOCKHASH':
        [target_wrapped] = self.stack.pop(1)
        [target] = self.stack.unwrap([target_wrapped])
        try:
          blockhash = web3.eth.get_block(target, full_transactions=False)['hash']
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [target_wrapped],
              "inputs": [target],
              "outputs": int(blockhash.hex() ,16),
            }
          )
          self.stack.append(int(blockhash.hex() ,16), op=opcode, source=tuple([target_wrapped]))
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [target_wrapped],
              "inputs": [target],
              "outputs": 0,
            }
          )
          self.stack.append(0, op=opcode, source=tuple([target_wrapped]))

      # get the coinbase of a block
      elif opcode == 'COINBASE':
        try:
          coinbase = web3.eth.get_block(web3.eth.block_number, full_transactions=False)['miner']
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": int(coinbase, 16),
            }
          )
          self.stack.append(int(coinbase, 16), op=opcode, source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, op=opcode, source=opcode)

      # get the timestamp of a block
      elif opcode == 'TIMESTAMP':
        try:
          timestamp = web3.eth.get_block(web3.eth.block_number, full_transactions=False)['timestamp']
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": timestamp,
            }
          )
          self.stack.append(timestamp, source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # get the number of a block
      elif opcode == 'NUMBER':
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": web3.eth.block_number,
            }
          )
          self.stack.append(web3.eth.block_number, source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # get difficulty of a block
      elif opcode == 'DIFFICULTY':
        try:
          difficulty = web3.eth.get_block(web3.eth.block_number, full_transactions=False)['difficulty']
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": difficulty,
            }
          )
          self.stack.append(difficulty, source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # get gaslimit of a block
      elif opcode == 'GASLIMIT':
        try:
          gasLimit = web3.eth.get_block(web3.eth.block_number, full_transactions=False)['gasLimit']
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": gasLimit,
            }
          )
          self.stack.append(gasLimit, source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # get the current EVM chain ID
      elif opcode == 'CHAINID':
        try:
          chainId = web3.eth.chain_id
          if chainId != None:
            self.stacktrace.append(
              {
                "pc": self.lastInstruction,
                "opcode": opcode,
                "wrapped": None,
                "inputs": None,
                "outputs": chainId,
              }
            )
            self.stack.append(chainId, source=opcode)
          else:
            self.stacktrace.append(
              {
                "pc": self.lastInstruction,
                "opcode": opcode,
                "wrapped": None,
                "inputs": None,
                "outputs": 0,
              }
            )
            self.stack.append(0, source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append(0, source=opcode)

      # get the balance of the current account
      elif opcode == 'SELFBALANCE':
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": web3.eth.get_balance( self.toAddress(self.address) ),
            }
          )
          self.stack.append(web3.eth.get_balance( self.toAddress(self.address) ), source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append( 0, source=opcode )

      # get the basefee of the network
      elif opcode == 'BASEFEE':
        try:
          baseFee = web3.eth.fee_history(1, web3.eth.block_number)['baseFeePerGas'][0]
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": baseFee,
            }
          )
          self.stack.append(baseFee, source=opcode)
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": None,
              "inputs": None,
              "outputs": 0,
            }
          )
          self.stack.append( 0, source=opcode )

      # remove the top item from the stack
      elif opcode == 'POP':
        [out_wrapped] = self.stack.pop(1)
        [out] = self.stack.unwrap([out_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [out_wrapped],
            "inputs": [out],
            "outputs": None,
          }
        )
        
      # load a word from memory at offset
      elif opcode == 'MLOAD':
        [pos_wrapped] = self.stack.pop(1)
        [pos] = self.stack.unwrap([pos_wrapped])
        loaded = int(self.memory.read(pos, 32).hex(), 16)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [pos_wrapped],
            "inputs": [pos],
            "outputs": loaded,
          }
        )
        self.stack.append(loaded, op=opcode, source=tuple([pos_wrapped]))

      # store a value to memory at offset
      elif opcode == 'MSTORE':
        [offset_wrapped, value_wrapped] = self.stack.pop(2)
        [offset, value] = self.stack.unwrap([offset_wrapped, value_wrapped])
        extended = bytearray.fromhex(self.logic.padHex(value, 64)[2:])
        self.memory.extend(offset, 32)
        self.memory.write(offset, 32, extended)
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [offset_wrapped, value_wrapped],
            "inputs": [offset, value],
            "outputs": None,
          }
        )

      # store an 8 byte value to memory at offset
      elif opcode == 'MSTORE8':
        [offset_wrapped, value_wrapped] = self.stack.pop(2)
        [offset, value] = self.stack.unwrap([offset_wrapped, value_wrapped])
        extended = bytearray.fromhex(self.logic.padHex(value, 64)[2:])
        self.memory.extend(offset, 1)
        self.memory.write(offset, 1, bytearray.fromhex(hex(extended[-1])[2:]))
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [offset_wrapped, value_wrapped],
            "inputs": [offset, value],
            "outputs": None,
          }
        )

      # store a value to storage at key
      elif opcode == 'SSTORE':
        [key_wrapped, value_wrapped] = self.stack.pop(2)
        [key, value] = self.stack.unwrap([key_wrapped, value_wrapped])
        self.storage[key] = value
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [key_wrapped, value_wrapped],
            "inputs": [key, value],
            "outputs": None,
          }
        )

      # load a value from storage at key
      elif opcode == 'SLOAD':
        [key_wrapped] = self.stack.pop(1)
        [key] = self.stack.unwrap([key_wrapped])
        try:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [key_wrapped],
              "inputs": [key],
              "outputs": self.storage[key],
            }
          )
          self.stack.append(self.storage[key], op=opcode, source=tuple([key_wrapped]))
        except:
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": [key_wrapped],
              "inputs": [key],
              "outputs": 0,
            }
          )
          self.stack.append(0, op=opcode, source=tuple([key_wrapped]))

      # change the program counter to destination
      elif opcode == 'JUMP':
        [dest_wrapped] = self.stack.pop(1)
        [dest] = self.stack.unwrap([dest_wrapped])
        try:
          if self.assembly[dest]['opcode']['name'] == "JUMPDEST":
            self.instruction = dest
            self.stacktrace.append(
              {
                "pc": self.lastInstruction,
                "opcode": opcode,
                "wrapped": [dest_wrapped],
                "inputs": [dest],
                "outputs": None,
              }
            )
        except:
          self.stacktrace.append(
              {
                "pc": self.lastInstruction,
                "opcode": opcode,
                "wrapped": None,
                "inputs": None,
                "outputs": None,
              }
            )
          
      # change the program counter to destination if the top item on the stack is zero
      elif opcode == 'JUMPI':
        [dest_wrapped, cond_wrapped] = self.stack.pop(2)
        [dest, cond] = self.stack.unwrap([dest_wrapped, cond_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [dest_wrapped, cond_wrapped],
            "inputs": [dest, cond],
            "outputs": None,
          }
        )
        if (self.assembly[dest]['opcode']['name'] == "JUMPDEST") and (cond != 0):
          self.instruction = dest
      
      # get the current program counter
      elif opcode == 'PC':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": self.instruction,
          }
        )
        self.stack.append(self.instruction, source=opcode)

      # get the current size of the EVM memory
      elif opcode == 'MSIZE':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": len(self.memory.getMemory().hex()),
          }
        )
        self.stack.append(len(self.memory.getMemory().hex()), source=opcode)
      
      # get the remaining gas of the current transaction
      elif opcode == 'GAS':
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": None,
            "inputs": None,
            "outputs": self.gas,
          }
        )
        self.stack.append(self.gas, source=opcode)

      # duplicate the nth item of the stack
      elif 'DUP' in opcode:
        dupn = int(opcode.replace('DUP', ''))
        args_wrapped = self.stack.pop(dupn)
        args = self.stack.unwrap(args_wrapped)
        input = copy.deepcopy(args)
        output = []
        args.insert(0, args[-1])
        args_wrapped.insert(0, args_wrapped[-1])
        for i in reversed(range(len(args))):
          output.append(args[i])
          self.stack.append_wrapped(args_wrapped[i])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [args_wrapped],
            "inputs": input,
            "outputs": output,
          }
        )

      # swap the nth item and the top item of the stack
      elif 'SWAP' in opcode:
        swapn = int(opcode.replace('SWAP', ''))
        args_wrapped = self.stack.pop(swapn+1)
        args = self.stack.unwrap(args_wrapped)
        input = copy.deepcopy(args)
        output = []
        temp_wrapped = args_wrapped[-1]
        temp = args[-1]
        args_wrapped[-1] = args_wrapped[0]
        args[-1] = args[0]
        args_wrapped[0] = temp_wrapped
        args[0] = temp
        for i in reversed(range(len(args))):
          output.append(args[i])
          self.stack.append_wrapped(args_wrapped[i])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [args_wrapped],
            "inputs": input,
            "outputs": output,
          }
        )

      # log an event with the given number of topics and data
      elif 'LOG' in opcode:
        try:
          logn = int(opcode.replace('LOG', ''))
          [offset_wrapped, size_wrapped] = self.stack.pop(2)
          [offset, size] = self.stack.unwrap([offset_wrapped, size_wrapped])
          rawTopics = []
          topics = []

          # if the number of topics is greater than 0, get the topics
          if logn > 0:
            rawTopics_wrapped = self.stack.pop(logn)
            rawTopics = self.stack.unwrap(rawTopics_wrapped)
          for topic in rawTopics:
            try:
              topics.append( self.logic.padHex(topic, 64) )
            except Exception as e:
              pass
          
          # read data from memory at offset and size
          mem = bytes(0)
          if len(self.memory.getMemory().hex()) > 0:
            mem = bytes(self.memory.read(offset, size))
            
          # log the event
          self.logEvent(
            {
              "signature": topics[0],
              "topicCount": logn,
              "topics": {
                "indexed": topics[1:],
                "unindexed": [mem.hex()[index : index + 64] for index in range(0, len(mem.hex()), 64)]
              },
            }
          )
          # calculate unindexed topics
          unindexed = []
          for n in range(math.floor(size / 32)):
            unindexed.append(offsetToMemoryName(offset + (n*32)))
        
          # add the log to the stacktrace
          self.stacktrace.append(
            {
              "pc": self.lastInstruction,
              "opcode": opcode,
              "wrapped": rawTopics_wrapped[1:],
              "event": {
                "signature": topics[0],
                "topicCount": logn,
                "topics": {
                  "indexed": topics[1:],
                  "unindexed": unindexed
                },
              }
            }
          )
        except Exception as e:
          pass
        
      # creates a new account with the given code from memory, account address as a hash of bytecode
      elif opcode == "CREATE":
        [value_wrapped, offset_wrapped, size_wrapped] = self.stack.pop(3)
        [value, offset, size] = self.stack.unwrap([value_wrapped, offset_wrapped, size_wrapped])
        addrBytes = bytes(bytearray.fromhex(self.toAddress(self.address)[2:]))
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [value_wrapped, offset_wrapped, size_wrapped],
            "inputs": [value, offset, size],
            "outputs": int((sha3(rlp.encode([addrBytes, 0]))[12:]).hex(), 16),
          }
        )
        self.stack.append( int((sha3(rlp.encode([addrBytes, 0]))[12:]).hex(), 16), op=opcode, source=tuple([value_wrapped, offset_wrapped, size_wrapped]) )

      # creates a new account with the given code from memory, account address as a hash of bytecode with salt
      elif opcode == "CREATE2":
        [value_wrapped, offset_wrapped, size_wrapped, salt_wrapped] = self.stack.pop(3)
        [value, offset, size, salt] = self.stack.unwrap([value_wrapped, offset_wrapped, size_wrapped, salt_wrapped])
        addrBytes = bytes(bytearray.fromhex(self.toAddress(self.address)[2:]))
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [value_wrapped, offset_wrapped, size_wrapped, salt_wrapped],
            "inputs": [value, offset, size, salt],
            "outputs": int((sha3(rlp.encode([addrBytes, 0]))[12:]).hex(), 16),
          }
        )
        self.stack.append( int((sha3(rlp.encode([addrBytes, 0]))[12:]).hex(), 16), op=opcode, source=tuple([value_wrapped, offset_wrapped, size_wrapped, salt_wrapped]) )

      # calls a contract with the given address, value, and data, and saves the return value to memory at retoffset
      elif opcode == 'CALL' or opcode == 'CALLCODE':
        [gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped] = self.stack.pop(7)
        [gas, address, value, argsOffset, argsSize, retOffset, retSize] = self.stack.unwrap([gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped],
            "inputs": [gas, address, value, argsOffset, argsSize, retOffset, retSize],
            "outputs": 1,
          }
        )
        self.stack.append(1, op=opcode, source=tuple([gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped]))

      # calls a contract with the given address, value, and data, and saves the return value to memory at retoffset
      # for delegatecalls, the call can modify the state of the caller
      elif opcode == 'DELEGATECALL' or opcode == 'STATICCALL':
        [gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped] = self.stack.pop(7)
        [gas, address, value, argsOffset, argsSize, retOffset, retSize] = self.stack.unwrap([gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped])
        self.stacktrace.append(
          {
            "pc": self.lastInstruction,
            "opcode": opcode,
            "wrapped": [gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped],
            "inputs": [gas, address, argsOffset, argsSize, retOffset, retSize],
            "outputs": 1,
          }
        )
        self.stack.append(1, op=opcode, source=tuple([gas_wrapped, address_wrapped, value_wrapped, argsOffset_wrapped, argsSize_wrapped, retOffset_wrapped, retSize_wrapped]))
      else:
        pass
    except Exception as e:
      self.stacktrace.append(
        {
          "pc": self.lastInstruction,
          "opcode": "ERROR",
          "inputs": None,
          "outputs": None,
        }
      )
      return (4, str(e), opcode)

  # executes operations on the EVM and returns a stacktrace of the operations
  def trace(self):
    try:
      instructionIndex = list(self.assembly.keys()).index(self.instruction)
      code = self.assembly[self.instruction]
    except:
      instructionIndex = list(self.assembly.keys()).index(str(self.instruction))
      code = self.assembly[str(self.instruction)]
    self.lastInstruction = self.instruction
    self.instruction = (list(self.assembly.keys())[instructionIndex+1])
    
    # executes the operation
    call = self.execute(code['opcode']['name'], code['argument'])
    
    # if the operation returns a value, execution stops
    if call != None:
      return ({
                "trace": self.stacktrace,
                "stack": self.getStack(),
                "logs": self.logs,
                "storage": self.storage,
                "memory": self.memory.getMemory().hex(),
                "returns": call
              })
    return self.trace()


import os
import pathlib
import datetime
import re
from alive_progress import alive_it

from .colors import colorLib

global logfile
logfile = f'{pathlib.Path(__file__).parent.parent.parent.resolve()}/logs/heimdall-{datetime.datetime.now().strftime("%Y%m%d%H%M%S")}.log'

# purge old log files
def purgeOldLogfiles():
  files = os.listdir(f'{pathlib.Path(__file__).parent.parent.parent.resolve()}/logs')
  files.remove('__init__.py')
  for file in files:
      if int(file.split('-')[1].replace('.log', '')) < int(datetime.datetime.now().strftime("%Y%m%d%H%M%S")) - 100000:
        os.remove(f'{pathlib.Path(__file__).parent.parent.parent.resolve()}/logs/{file}')
    # if timestamp is over 1 day old
      
# log to file, removing color codes
def logStripColor(file, type, log):
  with open(file,'a', encoding='UTF-8') as f:
    f.write(re.sub(r'(\\033|)\[\d*m', '', "{} {}\n".format(get_prefix(type), log)))

# log a traceback from traceback.format_exc()
def logTraceback(error, silent=False):
  traceback = error.strip().split("\n")
  errorLogString = f'Execution excepted: {traceback[-1]}'
  tracebackFiles = []
  for file in (traceback):
    if "File \"/" in file:
      tracebackFiles.append("".join(" , ".join(file.split(", ")[0:2]).split('"')[1:]))
  
  # parse traceback files in a nice tree
  for i, file in enumerate(tracebackFiles):
    errorLogString += f'\n{" "*29}{"├" if i+1 < len(tracebackFiles) else "└"}─({colorLib.RED}{i}{colorLib.RESET}) {file}'
  
  log("critical", errorLogString, silent)

# log a message with colored prefixes
def log(type, message, silent=False):
  if(type == "warning"):
    accent = colorLib.YELLOW
  elif(type == "critical"):
    accent = colorLib.RED
  elif(type == "success"):
    accent = colorLib.GREEN
  elif(type == "alert"):
    accent = colorLib.GREEN + colorLib.BOLD
  else:
    accent = colorLib.CYAN
  if not silent:
    print(datetime.datetime.now().strftime(f'[{accent}%H:%M:%S.%f{colorLib.RESET}] [{accent}{type.upper()}{colorLib.RESET}] {message}'))
  
  # send the log to the log file
  logStripColor(logfile, type, message)

# run a query to get a value, using the same log format
def query(type, default, message):
  if(type == "warning"):
    accent = colorLib.YELLOW
  elif(type == "critical"):
    accent = colorLib.RED
  elif(type == "success"):
    accent = colorLib.GREEN
  else:
    accent = colorLib.CYAN
  response = input(datetime.datetime.now().strftime(f'[{colorLib.CYAN}%H:%M:%S.%f{colorLib.RESET}] [{accent}{type.upper()}{colorLib.RESET}] {message}'))
  if len(response) < 1:
    return default
  return response

# gets the prefix for a log type
def get_prefix(type):
  if(type == "warning"):
    accent = colorLib.YELLOW
  elif(type == "critical"):
    accent = colorLib.RED
  elif(type == "success"):
    accent = colorLib.GREEN
  else:
    accent = colorLib.CYAN

  return(datetime.datetime.now().strftime(f'[{accent}%H:%M:%S.%f{colorLib.RESET}] [{accent}{type.upper()}{colorLib.RESET}]'))

# returns a styled progress bar given a list of items to create a progress bar for
def progress_bar(items, args):
  if args.verbose:
    return alive_it(items,title=get_prefix("info"), enrich_print=False, receipt=False, spinner=False, stats='(eta: {eta})')
  return items



from ..utils.colors import colorLib
from ..utils.version import getLocalVersion

# returns the heimdall global header
def getHeader():
  return (
    f'''{colorLib.BLUE}
 _   _  _____ ________  ________  ___   _      _     
| | | ||  ___|_   _|  \/  |  _  \/ _ \ | |    | |    
| |_| || |__   | | | .  . | | | / /_\ \| |    | |    
|  _  ||  __|  | | | |\/| | | | |  _  || |    | |    
| | | || |___ _| |_| |  | | |/ /| | | || |____| |____
\_| |_/\____/ \___/\_|  |_/___/ \_| |_/\_____/\_____/

{colorLib.RESET}Version:  {getLocalVersion()}
Author:   Jonathan Becker (jonathan@jbecker.dev)
GitHub:   https://github.com/Jon-Becker/heimdall

'''
  )


import json
import re
import pip_api
import requests

from ..config import *
from .logger import log
from .colors import colorLib

# Gets the latest version of eth-heimdall from PyPi
def getRemoteVersion():
  releaseRequest = requests.get('https://pypi.python.org/pypi/eth-heimdall/json', timeout=3)
  if releaseRequest.status_code == 200:
    latestVersionBody = json.loads(releaseRequest.text)
    return str(latestVersionBody['info']['version'].strip())
  return getLocalVersion()

# gets the local version of eth-heimdall
def getLocalVersion():
  try:
    return str(pip_api.installed_distributions(local=False)['eth-heimdall'].version)
  except:
    return str(getConfig()['build']['version'])
  
# gets the latest stable release of solidity from github
def getLatestSolidityRelease():
  try:
    releaseRequest = requests.get('https://api.github.com/repos/ethereum/solidity/releases', timeout=3)
    if releaseRequest.status_code == 200:
      request_body = json.loads(releaseRequest.text)
      return request_body[0]['tag_name'].replace("v", "^")
    return '>=0.8.0'
  except:
    return '>=0.8.0'
  
# performs a check to see if local version == remote version
def checkVersionUpToDate():
  remoteVersion = getRemoteVersion()
  localVersion = getLocalVersion()
  remoteVersionList = re.sub(r'[^0-9\.]' , '' , remoteVersion).split('.')
  localVersionList = re.sub(r'[^0-9\.]' , '' , localVersion).split('.')
  
  # Check each remote version according to SemVer and compare to local version
  isUpToDate = not ((remoteVersionList[0] > localVersionList[0]) or (remoteVersionList[1] > localVersionList[1]) or (remoteVersionList[2] > localVersionList[2]))
  
  return (isUpToDate, remoteVersion)

# performs a silent update of eth-heimdall through pip
def update(version):
  config = getConfig()
  if 'autoupdate' in config and config['autoupdate'] == True:
    log('info', f'Installing latest version {colorLib.CYAN + version[1] + colorLib.RESET}...')
    from subprocess import DEVNULL, STDOUT, check_call
    
    check_call(['python3', '-m', 'pip', 'install', 'eth-heimdall', '-U'], stdout=DEVNULL, stderr=STDOUT)

    if checkVersionUpToDate()[0]:
      log('success', f'Successfully updated to eth-heimdall {colorLib.GREEN + getLocalVersion() + colorLib.RESET}.')
    else:
      log('critical', 'Failed to install newest version of eth-heimdall from PyPi!')
  else:
    log('alert', f'You can update to version {colorLib.GREEN}{version[1]}{colorLib.RESET} by running: {colorLib.GREEN}pip install eth-heimdall=={version[1]} --U{colorLib.RESET} !')
    log('info', f'You can toggle autoupdates by running {colorLib.CYAN}heimdall config --autoupdate{colorLib.RESET}.')


from cmath import e
import os
import json
import time
import requests

from ..logger import log
from ..colors import colorLib
from ..io import outputDirectory, write, makePath

def fetchSourceCode(args, output, onlyAbi=False):
  log('info', 'Attempting to fetch source from EtherScan...')
  try:
    sourceRequest = requests.get(f'https://api.etherscan.io/api?module=contract&action=getsourcecode&address={args.target}')
    
    if sourceRequest.status_code == 200:
      
      # parse the response
      sourceBody = json.loads(sourceRequest.text)
      source = sourceBody['result'][0]['SourceCode'].replace("}}", "}", 1).replace("{{", "{", 1)
      abi = sourceBody['result'][0]['ABI']
      
      # write the abi to an output file
      log('info', f'Saving ABI file {colorLib.CYAN}{output.replace(os.getcwd(), ".")}/abi.json{colorLib.RESET}', True)
      try:
        abi = json.loads(abi)
        write(f'{output}/abi.json', json.dumps(abi, indent=2))
      except:
        log('critical', 'Writing ABI excepted! Defaulting to assembly builder.')
        return False
      
      if len(source) > 0:
        log('success', 'Found verified source on EtherScan!')

        # compiler language detection
        sourceType = "Solidity"
        sourceExtension = "sol"
        if "vyper" in sourceBody['result'][0]['CompilerVersion'].lower():
          sourceType = "Vyper"
          sourceExtension = "vy"
        log('info', f'Compiler language is {colorLib.CYAN}{sourceType}{colorLib.RESET}.', True)

        # multiple source files prokject
        if not onlyAbi:
          try:
            sourceObject = json.loads(source)
            try:
              log('info', 'Multiple file source detected!', True)
              sourcePath = makePath(f'{output}/source/')
              for key in sourceObject['sources']:
                log('info', f'Saving source file {colorLib.CYAN}{sourcePath.replace(os.getcwd(), ".")}{key}{colorLib.RESET}', True)
                dirPath = key.split("/")
                sourceName = dirPath.pop()
                dirpath = f'{sourcePath}{"/".join(dirPath)}'
                if not os.path.exists(dirpath):
                  os.makedirs(dirpath)
                write(f'{dirpath}/{sourceName}', sourceObject['sources'][key]['content'])
            except:
              log('critical', 'Fetching source excepted! Defaulting to assembly builder.')
              return False
          
          # single source file project
          except:
            log('info', 'Single file source detected!', True)
            write(f'{output}/source.{sourceExtension}', source)
            log('info', f'Saving source file {colorLib.CYAN}{output.replace(os.getcwd(), ".")}/source.{sourceExtension}{colorLib.RESET}')
          
        # all checks passed, return the abi
        log('success', f'Successfully retrieved contract source from EtherScan. Output saved to {colorLib.GREEN}{output.replace(os.getcwd(), ".")}{colorLib.RESET}.')
        return abi
        
      else:
        log('warning', 'No verified source found on EtherScan. Defaulting to assembly builder.')
        return False
    log('critical', 'Fetching source from EtherScan failed! Defaulting to assembly builder.')
    return False
  except:
    log('critical', 'Fetching source excepted! Defaulting to assembly builder.')
    return False
  
def fetchDeploymentBytecode(args, output):
  log('info', 'Attempting to fetch deployment bytecode from EtherScan...')
  try:
    sourceRequest = requests.get(f'https://api.etherscan.io/api?module=account&action=txlist&address={args.target}&startblock=0&endblock=99999999999&page=1&offset=1&sort=asc')
    
    if sourceRequest.status_code == 200:
      
      # parse the response
      sourceBody = json.loads(sourceRequest.text)
      
      # some sort of error encountered
      if "NOT" in sourceBody['message']:
        
        # rate limited, wait 5 seconds
        if "rate limit" in sourceBody['result']:
          log('warning', 'Etherscan rate-limited! Sleeping for 5 seconds.', True)
          time.sleep(5.01)
          return fetchDeploymentBytecode(args, output)
        else:
          
          # something else happened, return false
          log('critical', 'Couldn\'t fetch deployment bytecode from EtherScan.')
          return False
      else:
        if len(sourceBody['result']) > 0:
          return sourceBody['result'][0]['input']
                    
    log('critical', 'Fetching deployment bytecode from EtherScan failed!')
    return False
  except Exception as e:
    log('critical', 'Fetching deployment bytecode from EtherScan excepted!')
    return False


import os
import importlib

from ..modules.modules import getModules
from ..utils.colors import colorLib
from ..utils.version import getLocalVersion

# returns the heimdall help menu
def getHelp():
  
  modules = []
  fetched_modules = getModules()
  nl = "\n"
  
  modules.append(f'{colorLib.UNDERLINE}{colorLib.BOLD}{"#".rjust(9)}  |{"Name".rjust(fetched_modules[1]+2)}  |{"Description".rjust(fetched_modules[2]+2)}    {colorLib.RESET}')
  modules.append(f'{"".rjust(9)}  |{"".rjust(fetched_modules[1]+2)}  |{"".rjust(fetched_modules[2]+2)}')
  for item in fetched_modules[0]:
    modules.append(f'{str(len(modules)-2).rjust(9)}  |{item["title"].rjust(fetched_modules[1]+2)}  |{item["description"].rjust(fetched_modules[2]+2)}')
  
  return (
    f'''Usage: {colorLib.BOLD}heimdall MODULE [-v] [--default] [-arguments]{colorLib.RESET}

Powerful Ethereum smart contract toolkit for forensics, manipulation, and research.

Options:
  -h, --help                          Show the help message and exit
  -hh                                 Show advanced help message and exit
  --version                           Display version information and exit
  --update                            Updates heimdall to the latest release
  -v, --verbose                       Toggle verbose output

  Modules:
    Below is a list of modules currently supported on Heimdall

{nl.join(str(x) for x in modules)}

  Parameters:
    -m MODULE, --module MODULE        Operation module, either name or number from list
    -t TARGET, --target TARGET        Target of operation (file, transaction id,
                                        or address)
    -o PATH, --output PATH            Path to write output to
    -c ID, --chain ID                 Chain ID of target
    -p URL, --provider URL            URL of custom Ethereum provider

  Additional:
    --indent LEVEL                    Sets the indent level for output files (Default 2)
    --open, --edit                    Attempts to open nano / edit on the operation
    --redeploy ID                     Redeploys the contract from -n onto ID
    --beautify                        Attempts to beautify the downloaded contract using
                                        statistical renaming and spacing
    --default                         Always use defaults when prompted for input
    --flush, --ignore-cache           Flushes the cache and rewrites it
'''
  )


import os
import sys
import platform
import argparse
import pathlib
import datetime
import importlib
import traceback

from timeit import default_timer as timer

from .lib.utils.io import checksum, delete
from .lib.menus.help import getHelp
from .lib.utils.colors import colorLib
from .lib.menus.header import getHeader
from .lib.modules.modules import getModules
from .lib.utils.logger import log, logTraceback, logfile, purgeOldLogfiles
from .lib.utils.version import getLocalVersion, checkVersionUpToDate, update

def main(argv=None):
  
  # clear screen and print header
  purgeOldLogfiles()
  command = 'clear'
  if os.name in ('nt', 'dos'):
      command = 'cls'
  os.system(command)
  print(getHeader())
  
  heimdall = argparse.ArgumentParser(prog='heimdall', usage='heimdall [options]', add_help=False)

  # variable arguments
  heimdall.add_argument('-m', '--module', help="Operation module")
  heimdall.add_argument('-t', '--target', help="Operation target")
  heimdall.add_argument('-o', '--output', help="Path to write output to")
  heimdall.add_argument('-c', '--chain', help="Chain ID of target network")
  heimdall.add_argument('-p', '--provider', help="Custom provider URL")
  heimdall.add_argument('--indent', help="Sets the indent level for output files (Default 2)")
  heimdall.add_argument('--redeploy', help="Redeploy contract")

  # toggleable arguments
  heimdall.add_argument('-v', '--verbose', help="Use verbose mode", action="store_true")
  heimdall.add_argument('-h', '--help', help="Show the help menu", action="store_true")
  heimdall.add_argument('--version', help="Display version information", action="store_true")
  heimdall.add_argument('--update', help="Update heimdall to latest release", action="store_true")
  heimdall.add_argument('--beautify', help="Beautify contract, using statistical renaming", action="store_true")
  heimdall.add_argument('--default', help="Always use defaults when prompted for input", action="store_true")
  heimdall.add_argument('--flush', help="Flushes the cache", action="store_true")
  heimdall.add_argument('--autoupdate', help="Toggle autoupdates", action="store_true")
  heimdall.add_argument('--ignore-cache', help="Ignores the cache (SLOWER!)", action="store_true")
  heimdall.add_argument('--open', '--edit', help="Attempts to open nano / edit on the operation", action="store_true")  

  # parse arguments
  (args, extras) = heimdall.parse_known_args()
  
  # if theres an extra argument, it's assumed to be the module.
  # this will allow commands like `heimdall config` instead of `heimdall -m config`
  if len(extras) == 1:
    args.__setattr__('module', extras[0])
  
  log('debug', " ".join(sys.argv), True)
  log('debug', f'Uname: {platform.uname()}', True)
  log('debug', f'Checksum: {checksum(f"{pathlib.Path(__file__).parent.resolve()}/lib")}', True)
  log('debug', f'Heimdall Version: {getLocalVersion()}', True)
  
  startTime = timer()
  try:
    if args.help:
      print(getHelp())
      delete(logfile)
    else:
      if args.module:
        handled = False

        # get the selected module
        available_modules = getModules(args)
        if args.module.lower().isdigit() and (int(args.module) <= len(available_modules[0])-1):
          selected_module = importlib.import_module(available_modules[0][int(args.module)]['import'], package='heimdall')
          handled = True
        else:
          for module in available_modules[0]:
            if args.module.lower() == module["title"].lower():
              selected_module = importlib.import_module(module['import'], package='heimdall')
              handled = True
              break
        if not handled:
          log('critical', f'Module {colorLib.RED}{args.module}{colorLib.RESET} not found. Use -h to show the help menu.')
          
        # run the selected module
        try:
          log('debug', f'{selected_module.meta["title"]} ({selected_module.meta["version"]})', True)
          selected_module.main(args)
        except Exception as e:
          logTraceback(traceback.format_exc(), True)
          log('critical', f'Execution failed! Advanced logs available at {colorLib.RED + colorLib.UNDERLINE + logfile + colorLib.RESET} .')
                
        # run version check after module execution to minimize startup time 
        # from command send -> first log
        version = checkVersionUpToDate()
        if not version[0]:
          log('alert', f'This version of Heimdall is outdated!')
          update(version)
      else:
        log('critical', f'Missing a mandatory argument (-m or --module). Use -h to show the help menu.')
        delete(logfile)
        sys.exit(0)
    
      endTime = timer()
      log('info', f'Operation completed in {datetime.timedelta(seconds=(endTime-startTime))}.')
      if args.module in ('debug', 'config', 'publish'):
        delete(logfile)
    
  except KeyboardInterrupt:
    endTime = timer()
    log('critical', f'Operation aborted after {datetime.timedelta(seconds=(endTime-startTime))}.\n')
    sys.exit(0)
  
  except:
    logTraceback(traceback.format_exc(), True)
    sys.exit(0)
  
if __name__ == '__main__':
  sys.exit(main(sys.argv))


class Memory:
  def __init__(self):
    self.store = bytes(0)

  # safely calculate the cieling of a value to the uint256 maximum
  def safeCiel(self, value, cieling):
    r = value % cieling
    if r == 0:
      return value
    return value + cieling - r

  # extend memory by the given size after offset
  def extend(self, offset, size):
    if size == 0:
      return
    newSize = self.safeCiel(offset + size, 32)
    sizeDifference = newSize - len(self.store)
    if sizeDifference > 0:
      self.store = self.store + bytes(sizeDifference)

  # add a value to the memory at the given offset, with a fixed size.
  def write(self, offset, size, bufferval):
    if ( ( len(bufferval) == size ) and ( offset + size <= len(self.store) ) ):
      temp = bytearray.fromhex(self.store.hex())
      for i in range(size):
        temp[offset + i] = bufferval[i]
      self.store = bytes(temp)

  # read a value from the memory at the given offset, with a fixed size. 
  def read(self, offset, size):
    result = bytearray()
    temp = bytearray.fromhex(self.store.hex())
    loaded = temp[offset:offset+size]
    result += loaded
    
    # if memory length after offset is less than size, we append null bytes to the end of the memory.
    if len(loaded) < size:
      result += bytes(size-len(loaded))
    return result

  # returns the memory object
  def getMemory(self):
    return self.store



import os

from ..config import *
from ..utils.logger import *
from ..utils.io import readFile
from ..utils.colors import colorLib

meta = {
  "title": "Debug",
  "description": "Easily access Heimdall debug information",
  "author": "Jonathan Becker <jonathan@jbecker.dev>",
  "version": "v1.0.0",
}

# simply gets the most recent log file and gets its path
def main(args):
  try:
    max = 0
    for item in os.listdir(f'{pathlib.Path(__file__).parent.parent.parent.resolve()}/logs'):
      if "__" not in item and int(item.split("-")[1].replace(".log", "")) > max and item not in logfile:
        max = int(item.split("-")[1].replace(".log", ""))
        
    print(f"{colorLib.GREY + ''.join(readFile(f'{pathlib.Path(__file__).parent.parent.parent.resolve()}/logs/heimdall-{max}.log')) + colorLib.RESET}")
    log('debug', f'Latest log file available at: {colorLib.UNDERLINE+colorLib.CYAN}{pathlib.Path(__file__).parent.parent.parent.resolve()}/logs/heimdall-{max}.log{colorLib.RESET} .')
  except FileNotFoundError:
    log('info', f'No log files available.')


import os

from heimdall.lib.utils.io import write_config_value_at_path
from ..config import *
from ..utils.logger import *
from ..utils.colors import colorLib

meta = {
  "title": "Config",
  "description": "Easily modify the configuration on Heimdall",
  "author": "Jonathan Becker <jonathan@jbecker.dev>",
  "version": "v1.0.0",
}

def main(args):
  configPath = getConfigPath()
  config = getConfig()
  log('info', f'Heimdall configuration is located at: {colorLib.UNDERLINE+colorLib.CYAN+ configPath +colorLib.RESET} .')
  
  # handling for editing the config file in nano or edit
  if not args.open and not args.default:
    editFile = query('info', 'N', 'Would you like to edit the configuration file? (y/N):')
  elif args.default:
    editFile = "n"
    log('info', 'Would you like to edit the configuration file? (y/N): N')
  if args.open or editFile.lower() == "y":
    if os.name not in ('nt', 'dos'):
      os.system(f'nano {configPath}')
    else:
      os.system(f'edit {configPath}')
      
  # if they want to toggle autoupdate, set it in config
  if args.autoupdate:
    write_config_value_at_path(getConfigPath(), 'autoupdate', not getConfig()['autoupdate'])
    log('info', f"PyPi autoupdating is set to: {colorLib.CYAN}{'true' if getConfig()['autoupdate'] else 'false'}{colorLib.RESET}")


import json
import pathlib

from .utils.io import write, loadFileAsJson

def getConfig():
  try:
    return loadFileAsJson(f'{pathlib.Path(__file__).parent.parent.resolve()}/env/conf.json')
  except:
    # config file is missing, create it
    conf = {
      "build": {
        "version": "v1.0.3-stable"
      },
      "autoupdate": False,
      "defaults": {

        "providers": {
          "local": "http://127.0.0.1:7545",
          "remote": "https://mainnet.infura.io/v3/"
        }

      },

      "environment": {
        "apis": {
          "etherscan": ""
        }
      }
    }
    
    with open(getConfigPath(), 'w') as outfile:
      json.dump(conf, outfile, indent=4)
    return conf

# get the config file path, because it can be hidden in pypi directories
def getConfigPath():
  return f'{pathlib.Path(__file__).parent.parent.resolve()}/env/conf.json'


from hashlib import sha256

from web3 import Web3

from ..config import *
from ..utils.logger import log, query
from ..utils.colors import colorLib
from ..utils.io import outputDirectory, pathExists, readFile, write
from ..utils.eth.disassembler import disassemble
from ..utils.eth.builder import build
from ..utils.apis.etherscan import fetchSourceCode
from ..utils.eth.version import detectVersion, resolve

meta = {
  "title": "Decompile",
  "description": "Decompile and download the target smart contract",
  "author": "Jonathan Becker <jonathan@jbecker.dev>",
  "version": "v1.0.0",
}

def main(args):
  if args.target:
    target = args.target
    output = args.output

    # if the path exists, it's a local decompilation and the path contains bytecode
    args.local = pathExists(args.target)
    
    # its not a local file
    if not args.local:
      if not args.provider:
        args.provider = getConfig()['defaults']['providers']['remote']
        log('warning', f'Provider not set! Using default remote provider from configuration.')
       
      # try to connect to the web3 provider
      log('info', 'Establishing connection to provider...')
      if "http" in args.provider.lower():
        web3 = Web3(Web3.HTTPProvider(args.provider, request_kwargs={'timeout': 3}))
      elif "wss" in args.provider.lower():
        web3 = Web3(Web3.WebsocketProvider(args.provider, request_kwargs={'timeout': 3}))
      else:
        log('critical', f'Provider {colorLib.RED}{args.provider}{colorLib.RESET} doesn\'t seem valid.')
        return
      if web3.isConnected():
        log('success', 'Connection established!')
      else: 
        log('warn', f'Connection to {colorLib.YELLOW}{args.provider}{colorLib.RESET} failed!')
        return
      
      # fetch bytecode from web3
      log('info', f'Fetching target {colorLib.CYAN}{target}{colorLib.RESET}...')
      if not 'eth' in target:
        target = Web3.toChecksumAddress(target.lower())
      rawBytecode = web3.eth.get_code(target)
      bytecode = "".join(["{:02x}".format(v) for v in rawBytecode])
      
      # verify that the target is valid
      if len(bytecode) <= 1:
        log('warn', f'Target {colorLib.CYAN}{target}{colorLib.RESET} doesn\'t appear to be a contract address.')
        return
    else:
      # its a local file, we don't need to connect to web3. Fetch bytecode from the local file
      web3 = Web3()
      bytecode = readFile(args.target)[0]
      target = f'Local_0x{sha256(bytecode.encode("UTF-8")).hexdigest()[15:]}'
      args.target = "0x" + target
      
      # verify the bytecode is valid
      if not bytecode.startswith("0x") or " " in bytecode:
        log('critical', f'The file provided doesn\'t contain valid bytecode!')
        return
      bytecode = bytecode[2:]
      
    # create the output directory and save the target bytecode to it
    output = outputDirectory(output, target)
    write(f'{output}/bytecode.evm' , f'0x{bytecode}')
    
    # should we ignore decompilation completely and fetch source from etherscan?
    if not args.default:
      fetchFromEtherscan = query('info', "N", f'Would you like to attempt to fetch source from EtherScan? Otherwise, we will build source from assembly. [y/N]: ')
    else:
      fetchFromEtherscan = "N"
      log('info', f'Would you like to attempt to fetch source from EtherScan? Otherwise, we will build source from assembly. [y/N]: N')
    sourceFetched = False
    if fetchFromEtherscan == "Y":
      sourceFetched = fetchSourceCode(args, output)
    if sourceFetched == False:
      assembly = disassemble(bytecode, output, args)
      version = detectVersion(assembly, args)
      
      # log a warning if the version if unsupported
      log('info', f'Assembly heuristics suggest {colorLib.CYAN}{resolve(version[0][0])} <= EVM < {resolve(version[1][0])}{colorLib.RESET} and {colorLib.CYAN}Solidity >= {version[0][1]}{colorLib.RESET}.', True)
      if version[0][0] < 2:
        log('warning', f'Heimdall currently only has support for EVM versions after Constantinople. This contract may not decompile correctly.')
        
      # run the main decompilation build process
      log('info', 'Attempting to build source files from assembly...')
      source = build(assembly, args, output, web3)

  else:
    print('heimdall: error: Missing a mandatory option -t. Use -h to show the help menu.\n')
    return


from inspect import signature
import os
import sys

from numpy import sign

from bidict import bidict

from ..logger import log, progress_bar
from ..version import getLocalVersion
from ..colors import colorLib
from ..postprocess import postProcess
from ..version import getLatestSolidityRelease
from ..io import appendLine, write
from ..eth.functions import resolveFunctions

def build(assembly, args, output, web3, onlyAbi=False):
  sys.setrecursionlimit(5000)

  # resolves all found functions from evm assembly
  [abi, functions, indices, signatures, events, constantStorage, mappings] = resolveFunctions(assembly, args, output)
  
  # if we are not only building the abi, we build the contract solidity
  if not onlyAbi:
    sourcefile = f'{output}/decompiled.sol'
    indent_level = 0
    indent_increment = 2 if not args.indent else int(args.indent)
    
    # write the header to the decompiled file
    log('info', f'Creating source file {colorLib.CYAN}{sourcefile.replace(os.getcwd(), ".")}{colorLib.RESET}', True)
    source_header = f'''// SPDX-License-Identifier: MIT
pragma solidity {getLatestSolidityRelease()};

/// @title            Heimdall decompiled contract
/// @author           Jonathan Becker <jonathan@jbecker.dev>
/// @custom:version   {getLocalVersion()}
///
/// @notice           This contract was created using Heimdall toolkit's decompiler.
///                     It was generated directly by tracing the EVM opcodes from this contract.
///                     As a result, it may not compile or even be valid solidity code.
///                     Despite this, it should be obvious what each function does. Overall
///                     logic should have been preserved throughout decompiling.
///
/// @dev              Sometimes, return types don't match up with their originals. This
///                     shouldn't affect anything, but you should keep it in mind. The
///                     documentation goes into better detail on this.
/// @custom:github    You can find the open-source decompiler here
///                       https://github.com/Jon-Becker/heimdall
/// @custom:donations I spent ~250 hours on this program. It's open source and will always be
///                     free to use, so donations are always appreciated if you find it helpful.
///                     0x6666666b0B46056247E7D6cbdb78287F4D12574d   OR   jbecker.eth
'''
    write(sourcefile, source_header)
    lines = [
      f'contract Heim_{args.target[2:16]} {{',
      f'}}'
    ]

    # this will add a line to the source file after postprocessing each line
    def addline(line="", n=1, natspec=False):
      line = postProcess(line, signatures, events, constantStorage)
      for i in range(n):
        if natspec:
          lines.insert(-1, line)
        else:
          lines.insert(-1, line.replace("  ", " "))
    
    # write the resolved events to the decompiled file
    if any(interface['type'] == "event" for interface in abi):
      addline('''/// @notice         Contract event declaration
  /// @dev            If the signature wasn't resolved, parsing it is near-impossible. For now,
  ///                   we will default to bytes. TODO: indexed
  ''', natspec=True)
      
      # for each event in the abi, write the event declaration
      for event in [interface for interface in abi if interface['type'] == 'event']:
        inputs = []
        for input in event['inputs']:
          inputs.append(f'{input["type"]} {input["name"]}')
        inputs = str(tuple(inputs)).replace("'", "").replace(",)", ")")
        addline(f'event {event["name"]}{inputs};')
      addline(n=2)

    # write heimdall-specific constants and values to the decompiled file
    addline('''/// @notice           Heimdall constant declarations
  /// @dev              The below variables are used by Heimdall to mark panics, crashes,
  ///                     and other decompiler issues. These aren't present in the original
  ///                     contract, and are purely used for heuristics.
  
  string private constant EVM_PANIC = "Heimdall EVM panic! Report this with a GitHub issue.";
  
  /// @dev              Mappings that are dynamic but unresolved will be marked as generic.
  ///                     Please keep in mind multiple mappings may be marked as this generic
  ///                     mapping but still be separate mappings.
  
  mapping(bytes => mapping(bytes => bytes)) public mapping_generic;
  ''', natspec=True)
    
    # write all variables found to the decompiled file
    if any('constant' in interface and interface['constant'] == True for interface in abi):
      addline('''/// @notice           Contract variable declaration
  /// @dev              These variables have been detected because their getters
  ///                     were detected. These may or may not be in the constructor.
  ///                     Mappings will also appear here, if resolved.
  ''', natspec=True)

      # if we have mappings, we need to write them first
      if len(mappings) > 0:
        for mapping in mappings:
          addline(f'mapping({mapping["key"]} => {mapping["returns"]}) public mapping_{mapping["slot"]};')
        addline(n=1)
      
      # for each variable in the abi, write the variable declaration
      for constant in [interface for interface in abi if 'constant' in interface and interface['constant']]:
        outputs = []
        if "outputs" in constant:
          for output in constant['outputs']:
            outputs.append(f'{output["type"]} private ')
        if len(outputs) > 0:
          addline(f'{outputs[0]} _{constant["name"]};')
        else:
          addline(f'bytes private _{constant["name"]};')
      addline(n=2)
      
    # write the constructor to the decompiled file
    if any(interface['type'] == "constructor" for interface in abi):
      addline(f'constructor()  {{}}')
      addline(n=2)
    
    # write the functions to the decompiled file
    if any(interface['type'] == "function" for interface in abi):
      addline('''/// @notice         Contract function declaration
  /// @dev            Below is the function logic. As stated in the header, these will
  ///                   be *outlines* of the function logic and won't be exact.
  ///                   There are plans to improve this in the future.
  /// @notice         Additionally, view, pure, and payable detection are a bit buggy as of
  ///                   v1.0.0. There are plans to improve this detection.
  ''', natspec=True)

      # write getter functions before state modifying functions
      for getter in [interface for interface in abi if interface['type'] == 'function' and 'constant' in interface and interface['constant']]:
        outputs = []
        if "outputs" in getter:
          for output in getter['outputs']:
            outputs.append(f'{output["type"]} {"memory" if (output["type"] in ["string", "bytes"] or "[]" in output["type"]) else ""} {output["name"]}')
          output = str(tuple(outputs)).replace("'", "").replace(",)", ")")
        addline()
        if len(outputs) <= 0:
          
          # if we didn't find any outputs, we can't write a getter. Add a warning to the getter declaration and default to bytes
          addline('''/// @custom:warning No output found, reverting to bytes for this getter.''', natspec=True)
        addline(f'function {getter["name"]}() public '
                  f'{getter["stateMutability"] if getter["stateMutability"] != "nonpayable" else ""}'
                  f'{f" returns {output}" if len(outputs) > 0 else " returns (bytes memory ret0)"}'
                  f' {{')
        
        # write the getter logic
        for func in functions:
          if func.name == getter['name']:
            logic = sorted(func.logic, key=lambda x: x[0])
            for line in [x[1] for x in logic]:
              addline(line)
            break
        addline(f'return _{getter["name"]};')
        addline(f'}}')
      addline(n=2)
      
      # write state modifying functions after getter functions
      for interface in [interface for interface in abi if interface['type'] == 'function' and 'constant' in interface and not interface['constant']]:
        inputs = []
        for input in interface['inputs']:
          inputs.append(f'{input["type"]} {"calldata" if (input["type"] in ["string", "bytes"] or "[]" in input["type"]) else ""} {input["name"]}')
        inputs = str(tuple(inputs)).replace("'", "").replace(",)", ")")
        outputs = []
        if "outputs" in interface:
          for output in interface['outputs']:
            outputs.append(f'{output["type"]} {"memory" if (output["type"] in ["string", "bytes"] or "[]" in output["type"]) else ""} {output["name"]}')
          output = str(tuple(outputs)).replace("'", "").replace(",)", ")")
        addline()
        
        # write each functions warnings, if any
        for func in functions:
          if func.name == interface['name']:
            for line in func.warnings:
              addline(line, natspec=True)
      
        # add the function declaration
        addline(f'function {interface["name"]}{inputs} public '
                  f'{interface["stateMutability"] if interface["stateMutability"] != "nonpayable" else ""}'
                  f'{f" returns {output}" if len(outputs) > 0 else ""}'
                  f' {{')
        
        # write the function logic, removing the duplicate lines
        for func in functions:
          if func.name == interface['name']:
            logic = sorted(func.logic, key=lambda x: x[0])
            uniques = []
            for line in logic:
              if line[1] not in uniques:
                uniques.append(line[1])
            for line in uniques:
              addline(line)
            break
        addline(f'}}')

    # write each line to the decompiled file, accounting for indentation
    for line in progress_bar(lines, args):
      if line.startswith("}"):
        indent_level -= indent_increment
        
      appendLine(sourcefile, f'\n{" "*indent_level}{line}')
      
      if line.endswith("{"):
        indent_level += indent_increment
  
  log('success', f'Contract decompiled! Wrote source to {colorLib.GREEN}{sourcefile.replace(os.getcwd(), ".")}{colorLib.RESET}')
  return {
    "abi": abi
  }


class Any:
  pass

class Match:
  pass

import math
import re
import traceback
import numexpr

from time import time
from bidict import bidict
from eth_abi import decode_abi
from Crypto.Hash import keccak
from heimdall.lib.utils.logger import logTraceback
sha3 = lambda x: keccak.new(digest_bits=256, data=x).digest()

from .eth.opdict import opcodeDict, opArgNs

class Logic:
  def __init__(self):
    # set constants
    self.UINT256_MAX = 2 ** 256 - 1
    self.UINT255_MAX = 2 ** 255 - 1
    self.UINT256_CEILING = 2 ** 256
    self.UINT255_CEILING = 2 ** 255

  # safely sign an unsigned int
  def sign(self, unsigned):
    if ( unsigned & (1 << (255))) != 0:
      unsigned = unsigned - (1 << 256)
    return unsigned
  
  # safe add
  def add(self, a, b):
    return ((a + b) & self.UINT256_MAX)

  # safe multiplication
  def mul(self, a, b):
    if a == 0 or b == 0:
      return 0
    return ((a * b) & self.UINT256_MAX)

  # safe subtraction
  def sub(self, a, b):
    if a == b:
      return 0
    return ((b - a) & self.UINT256_MAX)

  # safe division
  def div(self, a, b):
    if a == 0 or b == 0:
      return 0
    return ((b // a) & self.UINT256_MAX)

  # safe signed division
  def sdiv(self, a, b):
    if a == 0 or b == 0:
      return 0
    a = self.sign(a)
    b = self.sign(b)
    flip = -1 if a * b < 0 else 1
    return flip * (abs(b) // abs(a))

  # safe modulo
  def mod(self, a, b):
    if a == 0:
      return 0
    return (b % a)

  # safe signed modulo
  def smod(self, a, b):
    a = self.sign(a)
    b = self.sign(b)
    flip = -1 if a < 0 else 1
    if a == 0:
      return 0
    return ( flip * (abs(b) % abs(a))) & self.UINT256_MAX

  # safe addmod operation
  def addmod(self, a, b, c):
    if a == 0:
      return 0
    return ( b + c ) % a

  # safe mulmod operation
  def mulmod(self, a, b, c):
    if a == 0:
      return 0
    return ( b * c ) % a

  # safe exponentiation op
  def exp(self, a, b):
    if a == 0:
      return 1
    if b == 0:
      return 0
    return pow(b, a, self.UINT256_CEILING)

  # pad a hex with null bytes to the given length
  def padHex(self, given_int, given_len):
      hex_result = hex(given_int)[2:]
      num_hex_chars = len(hex_result)
      extra_zeros = '0' * (given_len - num_hex_chars)

      return ('0x' + hex_result if num_hex_chars == given_len else
              '?' * given_len if num_hex_chars > given_len else
              '0x' + extra_zeros + hex_result if num_hex_chars < given_len else
              None)
  
  # calculate the least significant bit of a number
  def leastSignificantByte(self, value):
    if ( len(hex(value)) % 2 == 0 ) and ( value > 0 ):
      hexval = hex(value)
      return int(hexval[-2:],16)
    return 0

  # calculate the byte size of a value
  def byteSize(self, value):
    try:
      return math.ceil(len(hex(value)[2:])/2)
    except:
      return 0
  
  # convert a mask argument into a solidity type
  def resolveMask(args):
    arg = abs(args['val'])
    isPointer = args['isPointer']
    masks = bidict({
        ("bool", "bytes", "uint", "int", 'uint256', 'int256'): -1,
        ("uint8", "int8",): 8,
        ("uint16", "int16",): 16,
        ("uint32", "int32",): 32,
        ("uint64", "int64",): 64,
        ("uint128", "int128",): 128,
        ("address",): 160,
        ("uint256", "int256", "int", "uint", "address"): 256,
    })
    ret = []
    if arg in masks.inverse:
      ret += list(masks.inverse[arg])
    else:
      ret += ['address', 'uint256', 'int256', 'uint', 'int',]
    if arg % 2 == 0:
      bytesize = math.floor( ((arg/8)) )
      ret += [f"bytes{bytesize}", f"bytes{32-bytesize}", f"bytes"]
    
    # if this is used as a pointer, it could be an array
    if isPointer:
      ret += [ f"{potential}[]" for potential in ret]
    return ret

# convert a wrapepd stack value into a solidity logic line
def solidify_wrapped(_wrapped, vm, func=None):
  _ret = 'true'
  try:
    args = []

    # unwrap the value recursively
    def unwrap(_arg, args):
      if _match(_arg, (None, int, str)) or _match(_arg, (int, int, str)) or _match(_arg, (str, int, str)):
        if "PUSH" in _arg[2]:
          args.append(_arg[1])
        elif "CALLDATALOAD" in _arg[2]:
          args.append(f"arg{math.floor((_arg[0] - 4)/32)}")
        else:
          args.append(_arg[2])
      else:
        for arg in _arg:
          if type(arg) == tuple:
            unwrap(arg, args)
          elif type(arg) == str:
            args.append(arg)
    unwrap(_wrapped, args)
    
    # convert the unwrapped values into a solidity line
    _ret = solidity_operation(args, vm, func)
    
  except Exception as e:
    logTraceback(traceback.format_exc(), True)

  return _ret
    
# converts a list of unwrapped stack values into solidity by iterating over them to determine their solidity counterpart
def solidity_operation(_op, vm, func):
  operations = list(reversed(_op))
  while any(isinstance(op, (str,)) and op.lower() in opcodeDict.inverse for op in operations):
    for i, operation in enumerate(operations):
      if isinstance(operation, (str,)) and operation.lower() in opcodeDict.inverse:
        operations.pop(i)
        opInteger = opcodeDict.inverse[operation.lower()]
        opArgN = opArgNs[opInteger]
        opArgs = list(reversed([operations.pop(i-opArgN) for j in range(opArgN)]))
        [solidified, mem, mappings] = solidify(operation, vm, func, *opArgs)
        operations.insert(i-opArgN, solidified)
        break
  return operations[0]


def solidify(opcode, vm, func, *_args, mem={}, mappings={}):
  args = list(_args) + [None for x in range(7)]
    
  # if we run a keccak, save the value to a memory location for later use in mapping calculation
  if opcode == "SHA3":
    _rets = []
    mem['raw'] = sha3(vm.memory.read(0, 64)).hex()
    for n in range(math.floor(int(re.sub(r'[^0-9]', '', str(args[1]))) / 32)):
      _rets.append(offsetToMemoryName(n*32))
      mem[n] = {
        "value": vm.memory.read(n*32, 32).hex(),
        "source": _rets
      }
    return (f'keccak256(abi.encodePacked({" + ".join(_rets)}))', mem, mappings)
  
  # if we are loading from memory, we need to convert the offset into a memory name
  if opcode == "MLOAD":
    return (offsetToMemoryName(args[0]), mem, mappings)
  
  # if we are loading from storage and theres a keccak in the unwrapped stack, its most likely a mapping
  if opcode == "SLOAD" and any("keccak256" in str(arg) for arg in args):
    mappingSlot = int(mem[1]['value'], 16) if 1 in mem else 0
    if mappingSlot <= 32:
      
      # add mapping to the function mapping dict
      if mappingSlot not in func.mappings:
        func.mappings[mappingSlot] = {
          'slot': mappingSlot,
          'key': 'uint256',
          'returns': 'uint256',
        }
        
      if not isinstance(func.memory["var00"]["value"], (str)) and hex(func.memory["var00"]["value"]).startswith("0x4e487b71"):
        if isinstance(func.memlast["var00"]["value"], (str)):
          return (f'mapping_{mappingSlot}[{func.memlast["var00"]["value"]}]', mem, mappings)
        return (f'mapping_{mappingSlot}[EVM_PANIC]', mem, mappings)
      return (f'mapping_{mappingSlot}[{func.memory["var00"]["value"]}]', mem, mappings)
    elif m := resolveSlot(hex(mappingSlot)[2:]):
      
      # add mapping to the function mapping dict
      if m not in func.mappings:
        func.mappings[m] = {
          'slot': m,
          'key': 'uint256',
          'returns': 'uint256',
        }
        
      return (f'mapping_{m}[{func.memlast["var00"]["value"]}][{func.memory["var00"]["value"]}]', mem, mappings)
    else:
      return (f'mapping_generic[{func.memlast["var00"]["value"]}][{func.memory["var00"]["value"]}]', mem, mappings)

  # if the arg contains only arithmetic operations, evaluate them safely
  for i, arg in enumerate(args):
    if isArithmatic(str(arg)):
      try:
        args[i] = numexpr.evaluate(str(arg)).item()
      except: 
        args[i] = (2 ** 256 - 1)
  
  # handle variable casting to solidity type
  if opcode == 'AND':
    mask = False
    
    if isinstance(args[0], (int,)) or args[0].isnumeric():
      mask = int(args[0])
      temp = args[1]
    elif isinstance(args[1], (int,)) or args[1].isnumeric():
      mask = int(args[1])
      temp = args[0]
      
    if mask:
      # mask can either be the 0xFFFFFF kind or bitlength kind
      if re.match('(0x)?[fF]*', hex(mask)):
        castingType = Logic.resolveMask({"val": (len(hex(mask)[2:])*4), "isPointer": False})[0]
      else:
        castingType = Logic.resolveMask({"val": mask, "isPointer": False})[0]
      
      # make sure the variable isnt already cast to the type we are casting to
      isAlreadyCast = (re.match(r'^(int(\d+)?|uint(\d+)?|address|bytes(\d+)?|string|bool)\(.*?\)$', str(temp))) and (temp.split('(')[0] == castingType)
      
      if not isAlreadyCast: 
        return (f'{castingType}({temp})', mem, mappings)
      return (temp, mem, mappings)
  
  # constant solidity values for other opcodes
  solidified = {
    'ADD': f'{args[0]} + {args[1]}',
    'MUL': f'{args[0]} * {args[1]}',
    'SUB': f'{args[0]} - {args[1]}',
    'DIV': f'{args[0]} / {args[1]}',
    'SDIV': f'{args[0]} / {args[1]}',
    'MOD': f'{args[0]} % {args[1]}',
    'SMOD': f'{args[0]} % {args[1]}',
    'ADDMOD': f'({args[0]} + {args[1]}) % {args[2]}',
    'MULMOD': f'({args[0]} * {args[1]}) % {args[2]}',
    'EXP': f'{args[0]} ** {args[1]}',
    'SIGNEXTEND': f'SIGNEXTEND({args[0]}, {args[1]})',
    'LT': f'({args[0]}) < ({args[1]})',
    'GT': f'({args[0]}) > ({args[1]})',
    'SLT': f'({args[0]}) < ({args[1]})',
    'SGT': f'({args[0]}) > ({args[1]})',
    'EQ': f'({args[0]}) == ({args[1]})',
    'ISZERO': f'({args[0]}) == 0',
    'AND': f'{args[0]} & {args[1]}',
    'OR': f'{args[0]} | {args[1]}',
    'XOR': f'{args[0]} ^ {args[1]}',
    'NOT': f'~({args[0]})',
    'BYTE': f'({args[0]} >> (248 - {args[1]} * 8)) && 0xFF',
    'SHL': f'{args[0]} << {args[1]}',
    'SHR': f'({args[0]}) >> ({args[1]})',
    'SAR': f'{args[0]} >> {args[1]}',
    'SHA3': f'keccak256({args[0]})',
    'LOG0': f'emit {args[0]}()',
    'LOG1': f'emit {args[0]}({args[1]})',
    'LOG2': f'emit {args[0]}({args[1]}, {args[2]})',
    'LOG3': f'emit {args[0]}({args[1]}, {args[2]}, {args[3]})',
    'LOG4': f'emit {args[0]}({args[1]}, {args[2]}, {args[3]}, {args[4]})',
    'CREATE': f'// Contract appears to CREATE a new contract',
    'CREATE': f'// Contract appears to CREATE2 a new contract',
    'ADDRESS': 'address(this)',
    'BALANCE': f'address({args[0] if args[0] != None else "this"}).balance',
    'ORIGIN': 'tx.origin',
    'CALLER': 'msg.sender',
    'CALLVALUE': 'msg.value',
    'CALLDATASIZE': 'msg.data.length',
    'GASPRICE': 'tx.gasprice',
    'BLOCKHASH': f'blockhash({args[0] if args[0] != None else "block.number"})',
    'COINBASE': 'block.coinbase',
    'TIMESTAMP': 'block.timestamp',
    'NUMBER': 'block.number',
    'DIFFICULTY': 'block.difficulty',
    'GASLIMIT': 'block.gaslimit',
    'CHAINID': 'block.chainid',
    'SELFBALANCE': 'address(this).balance',
    'GAS': 'gasleft()',
    'SLOAD': f'LOAD{{{args[0]}}}',
    'RETURNDATASIZE': f'ext0.length'
  }
  
  try:
    return (solidified[opcode.upper()], mem, mappings)
  except Exception as e:
    return (opcode, mem, mappings)
  
# resolves a slot to a mapping
def resolveSlot(hex):
  for i in range(0, 0x20):
    if sha3(bytearray.fromhex(str(i).rjust(128, "0"))).hex() == hex:
      return i
  return False

# checks if a any list is in a list
def listContainsList(needleList, haystack):
  for needle in needleList:
    if needle in haystack:
      return True
  return False

# a recursive match function which can match dynamic expressions to patterns
def _match(expression, pattern):
  def _matcher(exp, pat, m):
    if isinstance(pat, str) and ":" in pat:
      setattr(m, pat.replace(":", ""), exp)
      return

    if pat is Any:
      return

    if isinstance(pat, (list, tuple)):
      if not isinstance(exp, (list, tuple)):
        raise ValueError()

      while True:
        if len(pat) == 0 and len(exp) == 0:
          return

        if pat[0] is Any:
          return

        _matcher(exp[0], pat[0], m)
        exp = exp[1:]
        pat = pat[1:]
    
    if isinstance(pat, type):
      if isinstance(exp, pat):
        return
      raise ValueError()
    
    if exp != pat:
      raise ValueError()
      

  try:
    match = Match()
    _matcher(expression, pattern, match)
  except Exception as e:
    return False
  return match

# determines the type of the given bytes, using length and decoding as heuristics
def determineType(func, retbytes):
  determined = "bytes"
  try:
    if type(retbytes) == int:
      retbytes = hex(retbytes)
    return_value = retbytes[2:]
    return_words = [return_value[index : index + 64] for index in range(0, len(return_value), 64)]
    
    # if theres more than one word, its an array or bytes or string
    if len(return_words) > 1:
      determined = "string"
    else:
      
      # its a single type, uint256, address, bool, or bytes32
      if retbytes == '0x0000000000000000000000000000000000000000000000000000000000000000':
        if func != None:
          func.warnings.append(f'/// @custom:warning This function returns either a boolean, uint, bytes32, or address.')
          func.warnings.append(f'///                   It\'s impossible to determine the exact return type.')
        determined = 'uint256'
      elif retbytes == '0x0000000000000000000000000000000000000000000000000000000000000001':
        if func != None:
          func.warnings.append(f'/// @custom:warning This function returns either a boolean, uint, bytes32, or address.')
          func.warnings.append(f'///                   It\'s impossible to determine the exact return type.')
        determined = 'bool'
      elif 20 <= len(retbytes[2:].replace("0", "", 64)) <= 40 :
        determined = 'address'
      else:
        if func != None:
          func.warnings.append(f'/// @custom:warning This function returns either a boolean, uint, bytes32, or address.')
          func.warnings.append(f'///                   It\'s impossible to determine the exact return type.')
        determined = 'uint256'
      
  except Exception as e:
    pass

  return determined

# some common types that we should override if possible
def commonTypes():
  return ('bytes', 'bytes[]', None)

# converting bytes to a given type, usually used for converting bytes to a string like revert reasons
def bytesToType(type, retbytes):
  try:
    return decode_abi(type, bytes.fromhex(retbytes[2:].replace('08c379a0', '')))
  except:
    return retbytes

# calculates a variable name from the offset provided
def offsetToMemoryName(offset): 
  try:
    return f'var{hex(math.floor(offset/4))[2:].ljust(2, "0")}'
  except:
    return offset
  
# ensures that s includes only the allowed characters for expressions.
def isArithmatic(s):  
  allowed_charset = "0123456789-+/*()%^<>!&|~. "
  return all(ch in allowed_charset for ch in s)


import os
import math
import json

from ..logger import log
from ..colors import colorLib
from ..io import outputDirectory, write
from ..eth.opdict import opcodeDict


def disassemble(bytecode, output, args):
  log('info', f'Beginning bytecode disassembly...')
  assemblyOut = []
  assembly = {}
  opIndex = 0
  totalBytes = len(str(math.floor(len(bytecode)/2)))
  hexBytes = len(hex(math.floor(len(bytecode)/2))[2:])

  # while theres still bytes to read, keep reading
  while opIndex < len(bytecode):
    opint = int(bytecode[opIndex:opIndex+2], 16)
    ophex = hex(opint)
    
    # if the opcode is not in the opcode dictionary, skip it
    if opint in opcodeDict:
      opcode = opcodeDict[ opint ].upper()
      pushVal = ""
      
      # if its a push opcode, we store the value too. Update the PC by the amount of bytes we read, plus the opcode length
      if "PUSH" in opcode:
        byteCount = int(opcode.replace("PUSH", ""))
        pushVal = bytecode[opIndex+2:(opIndex+2) + byteCount * 2 ]
        opIndex += byteCount*2

      # append the opcode to the assembly list
      assemblyOut.append(f'{hex(math.floor((opIndex)/2))[2:].zfill(hexBytes)} { str(math.floor((opIndex)/2)).zfill(totalBytes) } {opcode} {pushVal}')
      assembly[math.floor((opIndex)/2)] = {
        "instruction": math.floor((opIndex)/2),
        "opcode": {
          "hex": ophex,
          "int": opint,
          "name": opcode
        },
        "argument": int(pushVal, 16) if pushVal else ''
      }
    else:
      
      # append the opcode to the assembly list
      assemblyOut.append(f'{hex(math.floor((opIndex)/2))[2:].zfill(hexBytes)} { str(math.floor((opIndex)/2)).zfill(totalBytes) } unknown')
      assembly[math.floor((opIndex)/2)] = {
        "instruction": math.floor((opIndex)/2),
        "opcode": {
          "hex": ophex,
          "int": opint,
          "name": "unknown operation"
        },
        "argument": ""
      }

    opIndex += 2
  
  log('success', f'Disassembled {colorLib.GREEN}{str(math.floor(len(bytecode)/2))}{colorLib.RESET} bytes successfully.')
  write(f'{output}/assembly.asm', "\n".join(assemblyOut))
  return assembly


class Stack:
  def __init__(self, _stack=[]):
    self.stack = _stack
    self.trace = []

  # look at the next n values on the stack without modifying the stack
  def peek(self, n=1):
    res = []
    for i in range(1, n+1):
      res.append(self.stack[i-1:i][0])
    return res

  # remove the top n values from the stack
  def pop(self, n=1):
    res = []
    for i in range(n):
      res.append(self.stack.pop())
    return res

  # appends a wrapped value to the stack
  def append_wrapped(self, wrapped):
    self.stack.append(wrapped)

  # appends an operation to the stack
  def append(self, value, op=None, source=None):
    self.stack.append((op, value, source,))
    return self.stack

  # returns the wrapped value of an operation, which saves sources and operations performed on the stack
  def wrap(self, value, op=None, source=None):
    return (op, value, source, )

  # unwraps the `value` of an operation
  def unwrap(self, values):
    res = []
    for wrapped in values:
      res.append(wrapped[1])
    return res
    
  # insert an operation into the stack at the end of the stack
  def prepend(self, value):
    self.stack.insert(0, value)

  # returns the stack
  def getStack(self):
    return self.stack


from setuptools import Extension
from setuptools import setup

setup(name='eth_heimdall',
      version='1.0.3b8',
      description='Heimdall is an advanced and modular smart-contract toolkit which aims to make dealing with smart contracts on EVM based chains easier.',
      url='https://github.com/Jon-Becker/heimdall',
      entry_points={'console_scripts': ['heimdall=heimdall.__main__:main']},
      keywords=['ethereum', 'evm', 'decompiler', 'evm decompiler', 'smart contract', 'smart contract decompiler', 'evm smart contract decompiler', 'eth contract decompiler'],
      author='Jonathan Becker (jon-becker)',
      author_email='jonathan@jbecker.dev',
      license='MIT',
      
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Topic :: Software Development :: Disassemblers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Utilities'
      ],
      
      packages=[
        'heimdall',
        'heimdall/lib',
        'heimdall/logs',
        'heimdall/lib/menus',
        'heimdall/lib/modules',
        'heimdall/env',
        'heimdall/lib/utils',
        'heimdall/lib/utils/apis',
        'heimdall/lib/utils/eth',
        'heimdall/lib/utils/eth/classes'
        ],
      install_requires=[
          'requests',
          'web3',
          'numpy',
          'bidict',
          'alive-progress',
          'argparse',
          'argcomplete',
          'eth_abi',
          'eth_abi',
          'pip-api',
          'numexpr',
      ],
)


from ..logic import listContainsList

# resolve an integer to it's version name
def resolve(int):
  versions = ['Homestead', 'Byzantium', 'Constantinople', 'Istanbul', 'London', 'ArrowGlacier']
  return versions[int]

# detects the version of the compiler by looking at present opcodes and comparing them to the known versions
def detectVersion(assembly, args):
  versions = ['Homestead', 'Byzantium', 'Constantinople', 'Istanbul', 'London', 'ArrowGlacier']
  assemblyFlat = [assembly[pc]['opcode']['name'] for pc in assembly]
  min = [0, '0.1.0']
  max = [4, '0.8.14']
  
  if not listContainsList(['REVERT', 'RETURNDATASIZE', 'RETURNDATACOPY'], assemblyFlat):
    max[0] = 1
  elif not listContainsList(['SHL', 'SHR', 'SAR', 'EXTCODEHASH'], assemblyFlat):
    min[0] = 1
    max[0] = 2
  elif not listContainsList(['CREATE2', 'CHAINID', 'SELFBALANCE'], assemblyFlat):
    min[0] = 2
    max[0] = 3
  elif not listContainsList(['BASEFEE'], assemblyFlat): 
    min[0] = 3
    max[0] = 4
  
  if listContainsList(['BASEFEE'], assemblyFlat):
    max[1] = '0.8.14'
    min[1] = '0.8.7'
  elif listContainsList(['EXTCODEHASH'], assemblyFlat):
    max[1] = '0.5.0'
    min[1] = '0.4.12'
  elif listContainsList(['RETURNDATASIZE', 'RETURNDATACOPY', 'CREATE2', 'STATICCALL'], assemblyFlat):
    max[1] = '0.4.12'
    min[1] = '0.4.10'
  elif listContainsList(['REVERT'], assemblyFlat):
    max[1] = '0.4.10'
    min[1] = '0.4.7'
  elif listContainsList(['SHL', 'SHR', 'SAR'], assemblyFlat):
    max[1] = '0.4.7'

  return (min, max)